#-------------------------------------------------------------------------------
# Name:        McastSocket
# Purpose:
#
# Author:      LionKimbro
#
# Created:     19/12/2013
#
# Copyright 2014 LionKimbro
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of the GNU General Public License
#    as published by the Free Software Foundation; either version 2
#    of the License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
# This is based on https://wiki.python.org/moin/UdpCommunication
#-------------------------------------------------------------------------------

import socket
import struct


class McastSocket(socket.socket):
    """
    Class to make Multicast UDP handling easier.
    
    >>> recv_socket = McastSocket(local_port=5555, reuse=1)
    >>> recv_socket.mcast_add("235.0.0.1")
    >>> recv_socket.settimeout(3)
    >>> data, addr = recv_socket.recvfrom(2048)
    
    >>> recv_socket.sendto("hello", ("235.0.0.2", 8010))
    
    """

    def __init__(self, local_port=0, reuse=False):
        socket.socket.__init__(self, socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        if(reuse):
            self.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            if hasattr(socket, "SO_REUSEPORT"):
                self.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        self.bind(('', local_port))

    def mcast_add(self, addr, iface=socket.INADDR_ANY):
        """
        Add a multicast address to an interface
        
        :param addr: The multicast address to subscribe to
        :type addr: str
        :param iface: IP address of network interface on which to use this multicast address. Generally not required.
        :type iface: str
        """

        mreq = struct.pack("=4sl", socket.inet_aton(addr), socket.INADDR_ANY)
        self.setsockopt(
            socket.IPPROTO_IP,
            socket.IP_ADD_MEMBERSHIP,
            mreq)



import sys
sys.path.append("..")
import unittest
import AcraNetwork.Golay as Golay
from pstats import Stats
import cProfile
import random
import logging
import timeit

logging.basicConfig(level=logging.INFO)


class GolayTestCase(unittest.TestCase):

    def test_encode(self):
        g = Golay.Golay()
        for attempt in range(20):
            input = random.randint(0, 0xFFF)
            #print("{:0X}".format(g2.decode(g.encode(raw))))
            self.assertEqual(input, g.decode(g.encode(input)))

    def test_encode_string(self):
        g = Golay.Golay()
        v = 0x101
        s = g.encode(v, as_string=True)
        self.assertEqual(3, len(s))
        dec = g.decode(s)
        self.assertEqual(v, dec)

    def test_decode(self):
        g = Golay.Golay()
        v = 0x1007B408A722
        #print(repr(g.decode(v)))

    def test_with_error(self):
        g = Golay.Golay()
        for attempt in range(6):
            for error_bits in range(0, 7):
                logging.debug("----{}----".format(error_bits))
                input = random.randint(0, 0xFFF)
                encoded = g.encode(input)
                logging.debug("Raw={:#0X} Encoded={:#0X}".format(input, encoded))
                # Introduce errors
                bits_to_flip = random.sample(range(16), error_bits)
                for b in bits_to_flip:
                    encoded ^= (1 << b)
                logging.debug("Errorbits={} {:#0X}".format(bits_to_flip, encoded))
                decoded = g.decode(encoded)
                logging.debug("Decoded={:#X} Errors={}".format(g.decode(encoded), g._errors(encoded)))
                if error_bits < 4:
                    self.assertEqual(input, decoded)
                else:
                    self.assertNotEqual(input, decoded)



class GolayProfile(unittest.TestCase):

    def setUp(self):
        self.pr = cProfile.Profile()
        self.pr.enable()

    def tearDown(self):
        p = Stats(self.pr)
        p.sort_stats('cumtime')
        #p.print_stats()

    def test_profile(self):
        g = Golay.Golay()
        g2 = Golay.Golay()
        for val in range(100, 2000):
            #val = 100
            
            #print("{:0X}".format(g2.decode(g.encode(raw))))
            self.assertEqual(val, g2.decode(g.encode(val)))
        # Original implementation
        #         ncalls  tottime  percall  cumtime  percall filename:lineno(function)
        #         1    0.002    0.002    1.063    1.063 C:\ACRA\WORK\AXN_ENC_402\TICAD021\TICAD021\testbench\AcraNetwork\test\test_golay.py:69(test_profile)
        #     20    0.000    0.000    0.847    0.042 ..\AcraNetwork\Golay.py:70(decode)
        #     20    0.511    0.026    0.847    0.042 ..\AcraNetwork\Golay.py:121(_initgolaydecode)
        #     20    0.000    0.000    0.213    0.011 ..\AcraNetwork\Golay.py:53(encode)
        #     20    0.213    0.011    0.213    0.011 ..\AcraNetwork\Golay.py:42(_init_Table)
        # 276480    0.111    0.000    0.194    0.000 ..\AcraNetwork\Golay.py:106(_onesincode)
        # 276480    0.097    0.000    0.142    0.000 ..\AcraNetwork\Golay.py:94(_syndrome)
        # 276480    0.046    0.000    0.046    0.000 {method 'count' of 'str' objects}
        # 276500    0.045    0.000    0.045    0.000 ..\AcraNetwork\Golay.py:91(_syndrome2)
        # 276480    0.037    0.000    0.037    0.000 {built-in method builtins.bin}
        #     40    0.001    0.000    0.001    0.000 ..\AcraNetwork\Golay.py:37(__init__)

    def test_bits(self):
        error = 0x3
        size = 24
        for code in [0x0, 0x1, 0x37, 0xFFFFFF]:
            self.assertEqual(Golay.Golay._onesincode_old(code, size), Golay.Golay._onesincode(code, size))

        #print timeit.timeit('import AcraNetwork.Golay as Golay; Golay.Golay._onesincode(0x2, 24)', number=10000)
        #print timeit.timeit('import AcraNetwork.Golay as Golay; Golay.Golay._onesincode2(0x2, 24)', number=10000)

if __name__ == '__main__':
    unittest.main()



"""
.. module:: SimpleEthernet
    :platform: Unix, Windows
    :synopsis: A very trimmed down set of classes to unpack the common network packet formats

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
import socket
from functools import reduce
from zlib import crc32


def unpack48(x):
    """
    Unpack a 48bit string returning an integer

    :param x: 6 byte buffer
    :type x: bytes

    :rtype: int 
    """
    x2, x3 = struct.unpack('>HI', x)
    return x3 | (x2 << 32)


def mactoreadable(macaddress):
    """
    Convert a macaddress into the readable form
    
    :param macaddress: The mac address in integer format
    :type macaddress: int
    
    :rtype: str
    """
    mac_string = ""
    b = []
    for i in range(6):
        eachbyte = (macaddress >> i*8) & 0xFF
        b.append(eachbyte)

    return "{:02X}:{:02X}:{:02X}:{:02X}:{:02X}:{:02X}".format(b[5], b[4], b[3], b[2],  b[1], b[0])


def ip_calc_checksum(pkt):
    """
    Calculate the checksum of a packet
    
    :param pkt: The IP packet header packed into bytes
    :type pkt: str|bytes
    :return: 
    """

    if len(pkt) % 2 == 1:
        pkt += b"\0"
    s = sum(struct.unpack("<{}H".format(len(pkt) // 2), pkt))
    s = (s >> 16) + (s & 0xffff)
    s += s >> 16
    s = ~s
    return s & 0xffff


def combine_ip_fragments(packets):
    """
    Combine the lists of fragmented IP packets into one IP packet

    :type packets: list[IP]
    :rtype: IP
    """
    ident = None
    # Check first the we have the correct imput
    for packet in packets:
        if not isinstance(packet, IP):
            raise Exception("packet is not of type IP")
        if ident is not None:
            if packet.id != ident:
                raise Exception("All packets should have the same ID field")
        ident = packet.id

    # Create the IP packet we will return
    combined_ip = IP()
    combined_ip.flags = 0x0
    combined_ip.fragment_offset = 0x0
    combined_ip.payload = bytes()
    # Create the header from the first fragment. Order the packets via the fragment offsets
    # I don't verify the offsets but blindly combine them. TODO: Improve
    for idx, packet in enumerate(sorted(packets, key=lambda x: x.fragment_offset)):
        if idx == 0:
            for attr in ["srcip", "dstip", "protocol", "version", "ihl", "dscp", "id", "ttl"]:
                setattr(combined_ip, attr, getattr(packet, attr))
        # Add the payload
        combined_ip.payload += packet.payload

    # Return the IP packet
    return combined_ip


class Ethernet(object):
    """
    This is simple class to pack or unpack an Ethernet packet. Handles very basic packets that are used in FTI
    
    Read an Ethernet Packet from a pcap file
    
    >>> import AcraNetwork.Pcap as Pcap
    >>> p = Pcap.Pcap("test_input.pcap")
    >>> mypcaprecord = p[0]
    >>> e = Ethernet()
    >>> e.unpack(mypcaprecord.packet)
    >>> print e
    SRCMAC=00:18:F8:B8:44:54 DSTMAC=E0:F8:47:25:93:36 TYPE=0X800
    
    :type type: int
    :type srcmac: int
    :type dstmac: int
    :type payload: bytes
    
    """
    HEADERLEN = 14
    HEADERLEN_VLAN = 18
    TYPE_IP = 0x800
    TYPE_IPv4 = 0x800  #:(Object Constant) IPv4 Type Constant
    TYPE_IPv6 = 0x86DD  #:(Object Constant) IPv6 Type Constant
    TYPE_ARP = 0x806   #:(Object Constant) ARP Type Constant
    TYPE_PAUSE = 0x8808   #:(Object Constant) PAUSE Type Constant
    TYPE_VLAN = 0x8100

    def __init__(self, buf=None):

        """
        Create an Ethernet packet object. 

        :param buf: If a buffer is passed in to the init method, it will be unpacked as a Ethernet packet
        :type buf: bytes

        """
        self.type = Ethernet.TYPE_IP #: The Ethertype field. Assign using the TYPE_* constants. https://en.wikipedia.org/wiki/EtherType
        self.srcmac = None #: The Ethernet source MAC Address. This is encoded into a 48bit field. https://en.wikipedia.org/wiki/MAC_address
        self.dstmac = None #: The Ethernet destination MAC Address. This is encoded into a 48bit field. https://en.wikipedia.org/wiki/MAC_address
        self.payload = None #: The Ethernet payload. Typically an IP packet.
        self.vlan = False
        self.vlantag = 0xFFFF

        if buf is not None:
            self.unpack(buf)

    def unpack(self, buf, fcs=False):
        """
        Unpack a raw byte stream to an Ethernet object

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """

        self.dstmac = unpack48(buf[:6])
        self.srcmac = unpack48(buf[6:12])
        (self.type,) = struct.unpack_from('>H',buf,12)
        if fcs:
            self.payload = buf[Ethernet.HEADERLEN:-4]
            _fcs_buf = buf[-4:]
            exp_crc = crc32(buf[:-4])
            (act_crc,) = struct.unpack("I", _fcs_buf)
            if exp_crc != act_crc:
                raise Exception("FCS is wrong. Exo={:#0X} Act={:#0X}".format(exp_crc, act_crc))
        else:
            self.payload = buf[Ethernet.HEADERLEN:]
        return True

    def pack(self, fcs=False):
        """
        Pack the Ethernet object into a buffer
        
        :rtype: bytes
        """
        if self.dstmac == None or self.srcmac == None or self.type == None or self.payload == None:
            raise ValueError("All three required Ethernet fields are not complete")
        if self.vlan:
            header = struct.pack('>HIHIHHH',(self.dstmac>>32),(self.dstmac&0xffffffff),(self.srcmac>>32),(self.srcmac&0xffffffff), Ethernet.TYPE_VLAN, self.vlantag, self.type)
        else:
            header = struct.pack('>HIHIH',(self.dstmac>>32),(self.dstmac&0xffffffff),(self.srcmac>>32),(self.srcmac&0xffffffff), self.type)

        if fcs:
            _crc = crc32(header + self.payload) & 0xffffffff
            return header + self.payload + struct.pack("I", _crc)
        else:
            return header + self.payload

    def __repr__(self):
        return "SRCMAC={} DSTMAC={} TYPE={:#0X}".format(mactoreadable(self.srcmac), mactoreadable(self.dstmac), self.type)

    def __eq__(self, other):
        if not isinstance(other, Ethernet):
            return False

        for attr in ["type", "dstmac", "srcmac", "payload"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True


class IP(object):
    """
    Create or unpack an IP packet https://en.wikipedia.org/wiki/IPv4#Header
    
    If you wanted to unpack an Ethernet object payload which contains an IP packet
    
    >>> i = IP()
    >>> i.unpack(eth_pkt.payload)
    
    :type srcip: str
    :type dstip: str
    :type len: int
    :type flags: int
    :type protocol: int
    :type payload: bytes
    :type version: int
    :type ihl: int
    :type dscp: int
    :type id: int
    :type ttl: int
    
    """

    PROTOCOL_ICMP = 0x01  #:(Object Constant) ICMP Protocol Constant
    PROTOCOL_IGMP = 0x02  #:(Object Constant) IGMP Protocol Constant
    PROTOCOL_TCP = 0x6  #:(Object Constant) TCP Protocol Constant
    PROTOCOL_UDP = 0x11  #:(Object Constant) UDP Protocol Constant
    FLAG_DONT_FRAGMENT = 0x2
    FLAG_MORE_FRAGMENTS = 0x1

    PROTOCOLS = {"ICMP":PROTOCOL_ICMP, "IGMP" : PROTOCOL_IGMP, "TCP":PROTOCOL_TCP, "UDP":PROTOCOL_UDP}  #:(Object Constant) Protocols available
    IP_HEADER_FORMAT = '>BBHHBBBBHII'
    IP_HEADER_SIZE = struct.calcsize(IP_HEADER_FORMAT)

    def __init__(self, buf=None):
        """
        Create an IP packet object. Currently supports only IPv4
        
        :param buf: If a buffer is passed in to the init method, it will be unpacked as a IP packet
        :type buf: bytes
        
        """
        self.srcip = None #: Source IP Address
        self.dstip = None #: Destination IP Address
        self.len = None #: Total Length. This is calculated when packing the packet
        self.flags = 0x0 #: Three bit field identifying a flag
        self.fragment_offset = None #: Fragment offset
        self.protocol = IP.PROTOCOL_UDP #: The type of the payload
        self.payload = None #: The IPv4 payload
        self.version = 4 #: IP version field
        self.ihl = 5 #: Header length in 32 bit words
        self.dscp = 0 #: Differentiated Services Code Point
        self.id = 0 #: Identification Field
        self.ttl = 20 #: Time to Live. In practice the hop count.
		
        if buf is not None:
            self.unpack(buf)

    def unpack(self, buf):
        """
        Unpack a raw byte stream to an IP object

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """
        if len(buf) < IP.IP_HEADER_SIZE:
            raise ValueError("Buffer too short for to be an IP packet")
        (na1, self.dscp, self.len, self.id, self.flags, na3, self.ttl, self.protocol, checksum, self.srcip, self.dstip) \
            = struct.unpack_from(IP.IP_HEADER_FORMAT,buf)
        self.fragment_offset = (((self.flags & 0x1f) << 8) + na3) * 8 
        self.flags = self.flags >> 5
        self.version = na1 >> 4
        self.ihl = na1 & 0xf
        self.srcip = socket.inet_ntoa(struct.pack('!I',self.srcip))
        self.dstip = socket.inet_ntoa(struct.pack('!I',self.dstip))
        # Fill IP payload with number of bytes declared in header's length field, leaving any trailer behind (e.g. typically padding to reach 64bytes)
        self.payload = buf[IP.IP_HEADER_SIZE:self.len]

        return True

    def pack(self):
        """
        Pack the IP object into a buffer
        
        :rtype: bytes
        """

        for word in [self.dscp,self.id,self.flags,self.ttl,self.protocol,self.srcip,self.dstip]:
            if word is None:
                raise ValueError("All required IP payloads not defined")

        (srcip_as_int,) = struct.unpack('!I',socket.inet_aton(self.srcip))
        (dstip_as_int,) = struct.unpack('!I',socket.inet_aton(self.dstip))
        self.len = IP.IP_HEADER_SIZE+len(self.payload)
        header = struct.pack(IP.IP_HEADER_FORMAT, 0x45, self.dscp, self.len, self.id, self.flags, 0, self.ttl,
                             self.protocol, 0, srcip_as_int,dstip_as_int)
        checksum = ip_calc_checksum(header)
        header = header[:10] + struct.pack('H',checksum) + header[12:]
        return header + self.payload

    def __repr__(self):
        protocol = ""
        for p,v in IP.PROTOCOLS.items():
            if v == self.protocol:
                protocol = p
        return "SRCIP={} DSTIP={} PROTOCOL={} LEN={}".format(self.srcip, self.dstip, protocol, self.len)


class IPv6(object):
    """
    Create an IPv6 packet https://en.wikipedia.org/wiki/IPv6_packet

    :type version
    :type traffic_class
    :type flow_label
    :type len
    :type next_header
    :type hop_limit
    :type srcip
    :type dstip
    
    """

    PROTOCOL_ICMP = 0x01  #:(Object Constant) ICMP Protocol Constant
    PROTOCOL_IGMP = 0x02  #:(Object Constant) IGMP Protocol Constant
    PROTOCOL_TCP = 0x6  #:(Object Constant) TCP Protocol Constant
    PROTOCOL_UDP = 0x11  #:(Object Constant) UDP Protocol Constant
    FLAG_DONT_FRAGMENT = 0x2
    FLAG_MORE_FRAGMENTS = 0x1

    PROTOCOLS = {"ICMP":PROTOCOL_ICMP, "IGMP" : PROTOCOL_IGMP, "TCP":PROTOCOL_TCP, "UDP":PROTOCOL_UDP}  #:(Object Constant) Protocols available
    IP_HEADER_FORMAT = '>IHBBIIIIIIII'
    IP_HEADER_SIZE = struct.calcsize(IP_HEADER_FORMAT)

    def __init__(self, buf=None):
        """
        
        :param buf: If a buffer is passed in to the init method, it will be unpacked as a IP packet
        :type buf: bytes
        
        """
        self.version = 0x6 #: IP version field
        self.traffic_class = 0x00
        self.flow_label = 0x00000
        self.len = None #: Total Length. This is calculated when packing the packet
        self.next_header = 0x3B # 0x3B = No next header
        self.hop_limit = 0x01
        self.srcip = None #: Source IP Address
        self.dstip = None #: Destination IP Address
        self.payload = bytes()

        if buf is not None:
            self.unpack(buf)

    def pack(self):
        """
        Pack the IP object into a buffer
        
        :rtype: bytes
        """

        for word in [self.srcip,self.dstip]:
            if word is None:
                raise ValueError("All required IP payloads not defined")

        self.len = len(self.payload)
        header = struct.pack(IPv6.IP_HEADER_FORMAT, ((self.version<<28)+(self.traffic_class<<24)+self.flow_label), self.len, self.next_header, self.hop_limit, self.srcip>>96, (self.srcip>>64)&0xFFFFFFFF,(self.srcip>>32)&0xFFFFFFFF, self.srcip&0xFFFFFFFF, self.dstip>>96, (self.dstip>>64)&0xFFFFFFFF,(self.dstip>>32)&0xFFFFFFFF, self.dstip&0xFFFFFFFF)

        return header + self.payload

    def unpack(self, buffer):
        raise Exception("Not implemented")


    def __repr__(self):
        protocol = ""
        for p,v in IPv6.PROTOCOLS.items():
            if v == self.protocol:
                protocol = p
        return "SRCIP={} DSTIP={} PROTOCOL={} LEN={}".format(self.srcip, self.dstip, protocol, self.len)


class UDP(object):
    """
    Class to build and unpack a UDP packet
    
    https://en.wikipedia.org/wiki/User_Datagram_Protocol#Packet_structure
    
    Packet structure::
    
        -----2B----- -----2B----- -----2B----- -----2B----- --0-65527B----
        | SRC PORT  |  DEST PORT |   LENGTH   | CHECKSUM   | PAYLOAD
        ------------ ------------ ------------ ------------ --------------
    
    Create a UDP packet
    
    >>> u = UDP()
    >>> u.dstport = 5500
    >>> u.srcport = 4400
    >>> u.payload = struct.pack('B',0x5)
    >>> mypacket = u.pack()
    
    :type srcport: int
    :type dstport: int
    :type len: int
    :type payload: bytes
    """


    UDP_HEADER_FORMAT = '>HHHH'
    UDP_HEADER_SIZE = struct.calcsize(UDP_HEADER_FORMAT)

    def __init__(self, buf=None):

        self.srcport = None #: The UDP source port number
        self.dstport = None #: The UDP desitnation port number
        self.len = None #: The length of the UDP header and payload in bytes
        self.payload = None #: The UDP payload

        if buf is not None:
            self.unpack(buf)

    def unpack(self,buf):
        """
        Unpack a raw byte stream to a UDP object

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """

        if len(buf) < UDP.UDP_HEADER_SIZE:
            raise ValueError("Buffer too short to be a UDP packet")
        (self.srcport,self.dstport,self.len,checksum) = struct.unpack_from(UDP.UDP_HEADER_FORMAT,buf)
        self.payload = buf[UDP.UDP_HEADER_SIZE:]

        return True

    def pack(self):
        """
        Pack the UDP object into a buffer
        
        :rtype: bytes
        """

        if self.srcport == None or self.dstport == None or self.payload == None:
            raise ValueError("All UDP fields need to be defined to pack the payload")

        self.len = len(self.payload) + UDP.UDP_HEADER_SIZE
        return struct.pack(UDP.UDP_HEADER_FORMAT,self.srcport,self.dstport,self.len,0) + self.payload

    def __repr__(self):
        return "SRCPORT={} DSTPORT={}".format(self.srcport, self.dstport)


class AFDX(object):
    """
    This class will  unpack an AFDX packet
    
    """

    HEADERLEN = 14
    DSTMAC_CONST = 0x3000000
    SRCMAC_CONST = 0x20000
    MIN_PAYLOAD_LEN = 42

    def __init__(self, buf=None):
        raise Exception("No working")
        self.type =None
        self.networkID = None
        self.equipmentID = None
        self.interfaceID = None
        self.vlink = None

        self.payload = None
        self.sequencenum = None
        if buf is not None:
            self.unpack(buf)

    def unpack(self, buf):
        self.set_dstmac(buf[:6])
        self.unpacksrcmac(unpack48(buf[6:12]))

        (self.type,) = struct.unpack_from('>H',buf,12)
        self.payload = buf[AFDX.HEADERLEN:-1]
        self.sequencenum = struct.unpack('B',buf[-1])

    def unpacksrcmac(self, mac):
        srcconstantf = mac >> 24
        #if srcconstantf != AFDX.SRCMAC_CONST:
        #    raise ValueError('Expected constant field of {:#x} in SrcMac Address'.format(AFDX.SRCMAC_CONST))
        #(self.networkID,self.equipmentID,self.interfaceID) = struct.unpack_from('BBB',mac[:3])
        #self.interfaceID = self.interfaceID >> 5

    def set_dstmac(self, mac):
        (dstconstantf, vlink) = struct.unpack_from('>IH',mac)
        #if dstconstantf != AFDX.DSTMAC_CONST:
        #    raise ValueError('Expected constant field of {:#x} in DestMac Address'.format(AFDX.DSTMAC_CONST))
        self.vlink = vlink

    def pack(self):

        if len(self.payload) < AFDX.MIN_PAYLOAD_LEN:
            raise ValueError('Minimum Payload of {} bytes'.format(AFDX.MIN_PAYLOAD_LEN))

        afdx_header = struct.pack('>IHHBBBBH',AFDX.DSTMAC_CONST, self.vlink, (AFDX.SRCMAC_CONST >> 8), 0, self.networkID,
                                  self.equipmentID, (self.interfaceID << 5), self.type)

        packet = afdx_header + self.payload + struct.pack('>B',self.sequencenum)

        return packet

    def __eq__(self, other):
        if not isinstance(other, AFDX):
            return False

        for attr in ["type", "networkID", "interfaceID", "equipmentID", "vlink", "sequencenum", "payload"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True


class ICMP(object):
    """
    ICMP packets.
    """

    TYPE_REPLY = 0x0
    TYPE_UNREACHABLE = 0X1
    TYPE_REDIRECT = 0X5
    TYPE_REQUEST = 0X8

    def __init__(self):
        self.type = None
        self.code = None
        self.request_id = None
        self.request_sequence = None
        self.payload = bytes()

    def pack(self):
        """
        Pack an ICMP object into a buff
        :return:
        """
        for attr in ("type", "code", "request_id", "request_sequence"):
            if type(getattr(self, attr)) != int:
                raise ValueError("Attribute {} is not an integer".format(attr))

        _hdr_no_checksum = struct.pack(">BBHHH", self.type, self.code, 0, self.request_id, self.request_sequence)
        _icmp_checksum = ip_calc_checksum(_hdr_no_checksum + self.payload)
        _hdr = _hdr_no_checksum[:2] + struct.pack('H', _icmp_checksum) + _hdr_no_checksum[4:]
        return _hdr + self.payload

    def unpack(self, buffer):
        raise NotImplementedError("Not implemented")


MOD = 1 << 16


def ones_comp_add16(num1,num2):
    result = num1 + num2
    return result if result < MOD else (result+1) % MOD


class IGMPv3(object):
    """
    Simplified IGMP
    """
    TYPE_QUERY = 0x11
    TYPE_MEMBERSHIP_REPORT = 0x22
    TYPE_REC_CHG_TO_EXCL_MODE = 4
    TYPE_REC_MODE_IS_EXCLUDE = 2

    IP_ADDR_QUERY = "224.0.0.1"
    IP_ADDR_JOIN = "224.0.0.22"
    MAC_ADDR = {IP_ADDR_QUERY: 0x1005e000001, IP_ADDR_JOIN: 0x1005e000016}

    def __init__(self):
        pass

    @staticmethod
    def membership_query():
        """
        Return a membership query
        :return:
        """
        query_type = 0x11
        max_resp_time_2p4s = 0x18
        cksum = 0xecd3
        mc_addr = "0.0.0.0"
        sts = 0x2
        qqic = 0x20
        num_src = 0x0
        return struct.pack(">BBH", query_type, max_resp_time_2p4s, cksum) + \
                socket.inet_aton(mc_addr) + struct.pack(">BBH", sts, qqic, num_src)

    @staticmethod
    def join_groups(groups):
        """
        Join the specified groups
        :param groups:
        :return: bytes
        """
        if len(groups) == 1:
           mode = IGMPv3.TYPE_REC_CHG_TO_EXCL_MODE
        else:
            mode = IGMPv3.TYPE_REC_MODE_IS_EXCLUDE

        _nochecksum = struct.pack(">BBHHH", IGMPv3.TYPE_MEMBERSHIP_REPORT, 0, 0, 0, len(groups))
        for _g in groups:
            _nochecksum += struct.pack(">BBH", mode, 0, 0) + socket.inet_aton(_g)

        all_16b_wds = struct.unpack(f">{len(_nochecksum)//2}H", _nochecksum)
        checksum = ~reduce(ones_comp_add16, all_16b_wds) & 0xFFFF
        return _nochecksum[:2] + struct.pack(">H", checksum) + _nochecksum[4:]








import sys
sys.path.append("../")
import unittest
import AcraNetwork.Chapter7 as ch7
import AcraNetwork.SimpleEthernet as eth
import AcraNetwork.Pcap as pcap
import os
import copy
import struct
import logging
from pstats import Stats
import cProfile


THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def buf_generator(count, llp_count=0):
    running_count = 0
    while running_count < count:
        buf_len = os.urandom((running_count+1) * 128)
        running_count += 1
        if running_count <= llp_count:
            low_latency = True
        else:
            low_latency = False
        yield buf_len, low_latency


class TestCaseCh7(unittest.TestCase):

    def test_basic(self):
        ch7_pkt = ch7.PTFR()
        ch7_pkt.length = 132
        ch7_pkt.llp = 0
        remainder = bytes()
        while remainder == bytes():
            ch7_pd = ch7.PTDP()
            ch7_pd.content = ch7.PTDP_CONTENT_MAC
            ch7_pd.fragment = ch7.PTDP_FRAGMENT_COMPLETE
            ch7_pd.payload = os.urandom(58)
            remainder = ch7_pkt.add_payload(ch7_pd.pack())
            #print("Packet added")
        buf = ch7_pkt.pack()
        #print("Len Buf={} Len Rem={}".format(len(buf), len(remainder)))
        # Check the comparsion
        ch7_pkt_copy = copy.deepcopy(ch7_pkt)
        self.assertTrue(ch7_pkt==ch7_pkt_copy)

        ch7_unpack = ch7.PTFR()
        ch7_unpack.length = 132
        self.assertTrue(ch7_unpack.unpack(buf))
        self.assertTrue(ch7_unpack==ch7_pkt)

    def test_partial_fill(self):
        ch7_pkt = ch7.PTFR()
        ch7_pkt.length = 140
        ch7_pkt.llp = 0
        remainder = bytes()
        while remainder == bytes():
            ch7_pd = ch7.PTDP()
            ch7_pd.content = ch7.PTDP_CONTENT_MAC
            ch7_pd.fragment = ch7.PTDP_FRAGMENT_COMPLETE
            ch7_pd.payload = os.urandom(58)
            remainder = ch7_pkt.add_payload(ch7_pd.pack())
            #print("Packet added")
        buf = ch7_pkt.pack()
        ch7_unpack = ch7.PTFR()
        ch7_unpack.length = 140
        self.assertTrue(ch7_unpack.unpack(buf))
        #print(repr(ch7_pkt))
        #print(repr(ch7_unpack))
        self.assertEqual(52, len(remainder))
        self.assertTrue(ch7_unpack == ch7_pkt)


    def test_comparsion(self):
        ch7_pd = ch7.PTDP()
        ch7_pd.content = ch7.PTDP_CONTENT_MAC
        ch7_pd.fragment = ch7.PTDP_FRAGMENT_COMPLETE
        ch7_pd.payload = os.urandom(60)
        ch_pd_copy = copy.deepcopy(ch7_pd)
        self.assertTrue(ch7_pd==ch_pd_copy)


class TestGenerators(unittest.TestCase):

    def test_ptdp_generator(self):
        for ptdp_pkt in ch7.datapkts_to_ptdp(buf_generator(5)):
            #print(repr(ptdp_pkt))
            self.assertIsInstance(ptdp_pkt, ch7.PTDP)

    def test_ptfr_geenerator(self):
        remainder = bytes()
        for ptfr in ch7.datapkts_to_ptfr(buf_generator(5), ptfr_len=200):
            #print(repr(ptfr))
            ch7_pkt = ch7.PTFR()
            ch7_pkt.length = 800
            ch7_pkt.unpack(ptfr.pack())
            for (p, remainder, e) in ch7_pkt.get_aligned_payload(remainder):
                if p is None:
                    continue
                else:
                    #print(repr(p))
                    self.assertEqual(0, p.length % 128)



class TestRealEthernet(unittest.TestCase):

    def setUp(self):
        self.pr = cProfile.Profile()
        self.pr.enable()

    def tearDown(self):
        p = Stats(self.pr)
        p.sort_stats('cumtime')
        #p.print_stats()


    pkts_sent = []
    @staticmethod
    def eth_gen(count, low_latency_pkts=None, size_mult=128):
        """

        :param count:
        :param low_latency_pkts:
        :param size_mult:
        :rtype:  collections.Iterable[bytes, bool]
        """

        pf = pcap.Pcap(THIS_DIR+"/generated_eth.pcap", mode="w")
        r = pcap.PcapRecord()
        for i in range(1, count + 1):
            e = eth.Ethernet()
            e.dstmac = 0xABABABABABAB
            e.srcmac = 0xDCDCDCDCDCDC
            e.type = e.TYPE_IP
            ip = eth.IP()
            ip.srcip = "192.168.28.3"
            ip.dstip = "192.168.28.1"
            u = eth.UDP()
            u.dstport = i
            u.srcport = i
            if size_mult == 0:
                u.payload = struct.pack(">{}H".format(64), *([0xA5A5] * (64)))
            else:
                u.payload = struct.pack(">{}H".format(i * size_mult), *([0xA5A5] * (i * size_mult)))
            ip.payload = u.pack()
            e.payload = ip.pack()

            if low_latency_pkts is None:
                low_latency = False
            elif i in low_latency_pkts:
                low_latency = True
            else:
                low_latency = False
            r.setCurrentTime()
            r.payload = e.pack()
            pf.write(r)
            TestRealEthernet.pkts_sent.append(e.pack())
            yield e.pack(), low_latency
        pf.close()

    def test_eth_in_packets(self):
        pf = pcap.Pcap(THIS_DIR+"/gen.pcap", mode="w")
        r = pcap.PcapRecord()
        remainder = bytes()
        eth_p = bytes()
        pkt_count = 1
        for ptfr in ch7.datapkts_to_ptfr(TestRealEthernet.eth_gen(3), ptfr_len=200):
            ch7_pkt = ch7.PTFR()
            ch7_pkt.length = 200
            ch7_pkt.unpack(ptfr.pack())
            for (p, remainder, e) in ch7_pkt.get_aligned_payload(remainder):
                if p is None:
                    continue
                else:
                    eth_p += p.payload
                    if p.fragment == ch7.PTDP_FRAGMENT_COMPLETE or  p.fragment == ch7.PTDP_FRAGMENT_LAST:
                        r.setCurrentTime()
                        r.payload = eth_p
                        # Verify the size of the packets and it's probably good. I could unpack them here too
                        self.assertEqual(pkt_count*2*128+eth.UDP.UDP_HEADER_SIZE+eth.IP.IP_HEADER_SIZE+eth.Ethernet.HEADERLEN, len(eth_p))
                        pkt_count += 1
                        pf.write(r)
                        eth_p = bytes()


                    #self.assertEqual(0, p.length % 128)
                    #self.assertTrue(p.length <= 1024)
        pf.close()

    def test_eth_in_packets_low_latency(self):
        #logging.basicConfig(level=logging.DEBUG)
        pf = pcap.Pcap(THIS_DIR+"/captured_llp.pcap", mode="w")
        r = pcap.PcapRecord()
        remainder = bytes()
        eth_p = bytes()
        pkt_count = 1
        ptfr_idx = 0
        ptdp_idx = 0
        pkt_size_mult = 16
        for ptfr in ch7.datapkts_to_ptfr(TestRealEthernet.eth_gen(20, low_latency_pkts=[2, 4, 10], size_mult=8), ptfr_len=400):
            ch7_pkt = ch7.PTFR()
            ch7_pkt.length = 400
            ch7_pkt.unpack(ptfr.pack())
            #b = open("gen_llp_{}.bin".format(ptfr_idx), mode="wb")
            ptfr_idx += 1
            #b.write(ptfr.pack())
            #b.close()
            #print(repr(ptfr))
            for (p, remainder, e) in ch7_pkt.get_aligned_payload(remainder):
                if p is None:
                    continue
                else:
                    #print(repr(p))
                    eth_p += p.payload
                    if p.fragment == ch7.PTDP_FRAGMENT_COMPLETE or  p.fragment == ch7.PTDP_FRAGMENT_LAST:
                        if ptdp_idx == 0 :
                            self.assertEqual(p.low_latency, True)
                        r.setCurrentTime()
                        r.payload = eth_p
                        # Verify the size of the packets and it's probably good. I could unpack them here too
                        #self.assertEqual(2*64+eth.UDP.UDP_HEADER_SIZE+eth.IP.IP_HEADER_SIZE+eth.Ethernet.HEADERLEN, len(eth_p))
                        if eth_p not in TestRealEthernet.pkts_sent:
                            self.assertTrue(False)
                        pkt_count += 1
                        pf.write(r)
                        eth_p = bytes()
                    ptdp_idx += 1
        self.assertEqual(pkt_count, 11)


                    #self.assertEqual(0, p.length % 128)
                    #self.assertTrue(p.length <= 1024)
        pf.close()

    def test_llc(self):
        llc_count = 0
        fill_count = 0
        mac_count = 0
        logging.basicConfig(level=logging.WARN)
        remainder = b""
        for i in range(3, 6):
            f = open(THIS_DIR+"/ptfr_{}.bin".format(i), "rb")
            ptfr_data = f.read()
            f.close()

            ch7_pkt = ch7.PTFR()
            ch7_pkt.length = len(ptfr_data)
            eth_p = b""
            ch7_pkt.unpack(ptfr_data)
            #print repr(ch7_pkt)

            for (p, remainder, e) in ch7_pkt.get_aligned_payload(remainder):
                if p is not None:

                    if p.content != ch7.PTDP_CONTENT_FILL:
                        mac_count += 1
                        #print(repr(p))
                    if p.low_latency:
                        llc_count += 1
                    if p.content == ch7.PTDP_CONTENT_FILL:
                        fill_count += 1
        #print("{} {} {}".format(llc_count, fill_count, mac_count))
        self.assertEqual(4, llc_count)
        self.assertEqual(164, fill_count)
        self.assertEqual(4, mac_count)

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
===== 
Proxy iNetx or IENA packets to UDP 
===== 


"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__version__ = "0.1.0"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import sys
sys.path.append("..")
import socket
import argparse
import AcraNetwork.iNetX as inetx
import AcraNetwork.IENA as iena
import AcraNetwork.McastSocket as McastSocket

VERSION = __version__

parser = argparse.ArgumentParser(description='Proxy iNetX or IENA packets to UDP')
parser.add_argument('--ipaddress', type=str, default="235.0.0.1", required=True, help='The multicast IP address on which the iNetX or IENA packets are being transmitted')
parser.add_argument('--inetx', type=int, default=None, required=False, help='Receiving iNetX packets on this UDP port. Either this argument of --iena should be supplpied')
parser.add_argument('--iena', type=int, default=None, required=False, help='Receiving IENA packets on this UDP port')
parser.add_argument('--udp', type=int, default=None, required=True, help='Transmit UDP packets on this UDP port')
parser.add_argument('--version', action='version', version='%(prog)s {}'.format(VERSION))

args = parser.parse_args()

if not (args.inetx or args.iena):
    print(parser.print_help())
    sys.exit(1)

# The incoming iNetx port
if args.inetx:
    incoming_udp_port = args.inetx
else:
    incoming_udp_port = args.iena

# Outgoing UDP port
outgoing_udp_port = args.udp

#------------------------------------------------------------
# Setup a socket to recieve all traffic
#------------------------------------------------------------
try:
    recv_socket = McastSocket.McastSocket(local_port=incoming_udp_port, reuse=1)
    recv_socket.mcast_add(args.ipaddress, '0.0.0.0')
    recv_socket.settimeout(10)
except:
    print("Can't bind to socket {} on multicast {}".format(incoming_udp_port, args.ipaddress))
    exit()

tx_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

packet_count = 1


while True:
    # Capture some data
    try:
        data, addr = recv_socket.recvfrom(2048) # buffer size is 1500 bytes
    except socket.timeout:
        print("ERROR: No incoming packets received on UDP port {} on multicast {}. Timeout on socket".format(
            incoming_udp_port, args.ipaddress))
        exit()

    (udpsrcport,srcipaddr) = addr
    # Decode it as iNetx
    if args.inetx:
        avionics_packet = inetx.iNetX()
    else:
        avionics_packet = iena.IENA()

    try:
        avionics_packet.unpack(data)
    except ValueError:
        # This isn't an inetx packet
        continue
    else:
        packet_count += 1
        # Transmit the _payload back out
        tx_socket.sendto(avionics_packet.payload, (args.ipaddress, outgoing_udp_port))
        # Print some info for the user
        if packet_count % 50 == 0:
            print(".")
        else:
            print(".",)


"""
.. module:: IENA
    :platform: Unix, Windows
    :synopsis: Class to pack and unpack IENA packets

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
import datetime,time
from collections import namedtuple


class IENA (object):
    """
    Class to :meth:`IENA.pack` and :meth:`IENA.unpack` IENA payloads. 
    
    IENA is an proprietary payload format
    developed by Airbus for use in FTI networks. It is usually transmitted in a UDP packet
    containing parameter data acquired from sensors and buses::
    
        ---2B--- ---2B-- -----------6B----------- -1B- -1B- ---2B--- ------0B to 65490B----- ---2B---
        | KEY   | SIZE  | TIME SINCE START YR(US)| ST | N2 | SEQNUM |      PARAMETERS       |   END  |
        -------- ------- ------------------------ ---- ---- -------- ----------------------- --------  
    
    Create an IENA packet and return the packed buffer
    

    >>> i = IENA()
    >>> i.key = 0xDC
    >>> i.sequence = 1
    >>> i.endfield = 0xDEAD
    >>> i.keystatus = 0
    >>> i.status = 0x0
    >>> i.timeusec = int(10e6)
    >>> i.payload = struct.pack('H',0x5)
    >>> i.pack()
    b'\\x00\\xdc\\x00\\t\\x00\\x00\\x00\\x98\\x96\\x80\\x00\\x00\\x00\\x01\\x05\\x00\\xde\\xad'
    
    Read in some data stored in a UDP packet in a pcap file
    
    >>> import AcraNetwork.Pcap as pcap
    >>> p = pcap.Pcap("../test/iena_test.pcap")
    >>> rec_payload = p[0].payload
    >>> i = IENA()
    >>> i.unpack(rec_payload[0x2a:])  # Offset into the pcap record
    True
    >>> print("{:#0X}".format(i.key))
    0X1A

    :type key: int
    :type size: int
    :type timeusec: int
    :type keystatus: int
    :type status: int
    :type sequence: int
    :type endfield: int
    :type payload: bytes
     
    """
    IENA_HEADER_FORMAT = '>HHHIBBH'
    IENA_HEADER_LENGTH = struct.calcsize(IENA_HEADER_FORMAT)
    TRAILER_LENGTH = 2

    REQ_ATTR = ("key", "timeusec", "keystatus", "status", "sequence", "endfield", "payload")

    def __init__(self):

        self._key = None # know as ienakey
        self.size = None #: IENA packet size incl header. This is calculated automatically when packing the message
        self.timeusec = 0 #: Time of the first byte in the payload in us since Jan 1st of the current year
        self.keystatus = None  #: Key Status. Fully described in the IENA standard
        self.status = None #: N2 Status. Fully described in the IENA standard
        self.sequence = None #: Sequence Number. Circular counter unique per key. Wraps at 16 bits.
        self.endfield = 0xdead #: Trailer field in the IENA packet
        self.payload = None #: Payload of the IENA packet

        self._packetStrut = struct.Struct(IENA.IENA_HEADER_FORMAT)
        # only calculate this once TODO: This is wrong
        self._startOfYear = datetime.datetime(datetime.datetime.today().year, 1, 1, 0, 0, 0,0)
        self.lengthError = False # Flag to verify the buffer length

        # The required attributes
        self._req_attr = IENA.REQ_ATTR

    @property
    def key(self):
        """
        The IENA Key. Identifies the IENA packet type
        """
        return self._key

    @key.setter
    def key(self,key):
        self._key = key

    @property
    def streamid(self):
        """
        Alias of the IENA key.
        """
        return self._key

    @streamid.setter
    def streamid(self, key):
        self._key = key

    @property
    def n2(self):
        """
        Alias of the N2 Status.
        """
        return self.status

    @n2.setter
    def n2(self, n2):
        self.status = n2

    def unpack(self, buf):
        """
        Unpack a raw byte stream to an IENA object
        Accepts a buffer to unpack as the required argument
        
        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """

        # Some checking
        if len(buf) < IENA.IENA_HEADER_LENGTH:
            raise ValueError("Buffer passed to unpack is too small to be an IENA packet")

        (self.key, self.size, timehi, timelo, self.keystatus, self.status, self.sequence)  = self._packetStrut.unpack_from(buf)
        self.timeusec = timelo + timehi * 2**32

        if self.size*2 != len(buf):
            raise Exception("Length field does not match the size of the packet")

        self.payload = buf[IENA.IENA_HEADER_LENGTH:-2]
        (self.endfield,) = struct.unpack_from(">H",buf, -2) # last two bytes are the trailer

        return True

    def pack(self):
        """
        Pack the IENA payload into a binary string
        
        :rtype: bytes
        """

        for attr in self._req_attr:
            if getattr(self, attr) is None:
                raise ValueError("Require {} is not defined".format(attr))

        timehi = self.timeusec >> 32
        timelo = self.timeusec % 0x100000000

        self.size =  (len(self.payload)  + IENA.IENA_HEADER_LENGTH + IENA.TRAILER_LENGTH) // 2 # size is in words

        packetvalues = (self.key, self.size, timehi, timelo, self.keystatus, self.status, self.sequence)
        packet = self._packetStrut.pack(*packetvalues) + self.payload + struct.pack('>H', self.endfield)

        return packet

    def _getPacketTime(self):
        """
        Return the Packet time in standard UNIX time
        
        :rtype: int
        """
        return int(self.timeusec/1e6 + time.mktime(self._startOfYear.timetuple()))

    def setPacketTime(self, utctimestamp, microseconds=0):
        """ 
        Set the packet timestamp
        
        :param timestamp: The seconds timestamp based on he standard timeformat 
        :type timestamp: int
        :param microseconds: The microseconds part of the timestamp
        :type microseconds: int
        
        """
        seconds_this_year = utctimestamp - int(time.mktime(self._startOfYear.timetuple()))
        packettime = microseconds + int(seconds_this_year)*1000000
        self.timeusec = packettime

    def __repr__(self):
        return "IENAP: KEY={:#0X} SEQ={} TIMEUS={}".format(self.key, self.sequence, self.timeusec)

    def __eq__(self, other):

        for attr in self._req_attr:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __len__(self):
        return len(self.pack())


class MParameter(namedtuple("MParameter", "paramid, delay, dataset")):
    """
    The MParameter is a object representing each MParameter in an IENA-M packet

    :param paramid: The param ID for this parameter
    :type paramid: int
    :param delay: The delay for this parameter
    :type delay: int
    :param dataset: The dataset payload as a string
    :type dwords: bytes
    """

    def __repr__(self):
        return "ParamID={:#0X} Delay={} Dataset Length={}".format(self.paramid, self.delay, len(self.dataset))


class IENAM(IENA):
    """
    Support for IENA-M packets. Message Parameters with delay field
    
    This payload includes a defined parameter ID, a delay and a length field
    
    All this is encapsulated inside an IENA packet as defined above::
    
        ---2B--- ---2B-- ---2B--- ------0B to 65490B----------
        |PARAM  | DELAY | LENGTH | Dataset with opt 1B pad    |
        -------- ------- -------------------------------------  
    
    Unpack some received packet from the network
    
    >>> import socket
    >>> recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    >>> data, addr = recv_socket.recvfrom(2048)
    >>> i = IENAM()
    >>> i.unpack(data)
    >>> for param in i:
    ...   print param.paramid
    6
    
    """

    _FORMAT_ = ">HHH"
    _FORMAT_LEN_ = struct.calcsize(_FORMAT_)
    REQ_ATTR = ("key", "timeusec", "keystatus", "status", "sequence", "endfield", "payload", "parameters")

    def __init__(self):
        IENA.__init__(self)
        self.parameters = []  #: The list of all MParameters in thie IENA-M packet Each entry is of class :class:`MParameter`
        self._req_attr = IENAM.REQ_ATTR

    def unpack(self, buf):
        """
        Unpack a raw byte stream to an IENA-M object
        Accepts a buffer to unpack as the required argument

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """

        super(IENAM, self).unpack(buf)
        remaining_payload = self.payload

        while len(remaining_payload) > 0:
            (_paramid, _delay, datasetlength) = struct.unpack(
                IENAM._FORMAT_, remaining_payload[:IENAM._FORMAT_LEN_])
            # Check if we have enough payload
            if len(self.payload[IENAM._FORMAT_LEN_:]) < datasetlength:
                raise Exception("M Param dataset length {} larger than payload{}".format(
                    datasetlength, len(remaining_payload[IENAM._FORMAT_LEN_:])))

            mparam = MParameter(paramid=_paramid, delay=_delay,
                                dataset=remaining_payload[IENAM._FORMAT_LEN_:IENAM._FORMAT_LEN_+datasetlength])
            self.parameters.append(mparam)
            if datasetlength % 2 == 1:
                padding = 1
            else:
                padding = 0
            remaining_payload = remaining_payload[IENAM._FORMAT_LEN_+datasetlength+padding:]

    def pack(self):
        """
        Pack the IENA-M payload into a binary string

        :rtype: bytes
        """
        self.payload = b""
        for mparam in self:
            datasetlength = len(mparam.dataset)
            self.payload += struct.pack(IENAM._FORMAT_, mparam.paramid, mparam.delay, datasetlength)
            self.payload += mparam.dataset
            if datasetlength % 2 == 1:
                self.payload += struct.pack(">B", 0)

        return super(IENAM, self).pack()

    def __repr__(self):
        txt= "IENAM: KEY={:#0X} SEQ={} TIMEUS={} NUM_MPARAM={}\n".format(self.key, self.sequence, self.timeusec,
                                                                         len(self))
        for idx, param in enumerate(self):
            txt += " M-Param #{}:{}\n".format(idx, repr(param))

        return txt

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.parameters):
            _param = self.parameters[self._index]
            self._index += 1
            return _param
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.parameters)

    def __getitem__(self, key):
        return self.parameters[key]


class QParameter(namedtuple("QParameter", "paramid, dataset")):
    """
    The QParameter is a object representing each QParameter in an IENA-Q packet

    :param paramid: The param ID for this parameter
    :type paramid: int
    :param dataset: The dataset payload as a string
    :type dwords: bytes
    """

    def __repr__(self):
        return "ParamID={:#0X} Dataset Length={}".format(self.paramid, len(self.dataset))


class IENAQ(IENA):
    """
    Support for IENA-Q packets. Message Parameters without delay field

    This payload includes a defined parameter ID and a length field

    All this is encapsulated inside an IENA packet as defined above::

        ---2B---  ---2B--- ------0B to 65490B----------
        |PARAM   | LENGTH | Dataset with opt 1B pad    |
        --------  ------- -----------------------------

    Unpack some received packet from the network

    >>> import socket
    >>> recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    >>> data, addr = recv_socket.recvfrom(2048)
    >>> i = IENAQ()
    >>> i.unpack(data)
    >>> for param in i:
    ...   print param.paramid
    6

    """

    _FORMAT_ = ">HH"
    _FORMAT_LEN_ = struct.calcsize(_FORMAT_)
    REQ_ATTR = ("key", "timeusec", "keystatus", "status", "sequence", "endfield", "payload", "parameters")

    def __init__(self):
        IENA.__init__(self)
        self.parameters = []  #: The list of all QParameters in thie IENA-Q packet Each entry is of class :class:`QParameter`
        self._req_attr = IENAM.REQ_ATTR

    def unpack(self, buf):
        """
        Unpack a raw byte stream to an IENA-Q object
        Accepts a buffer to unpack as the required argument

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """

        super(IENAQ, self).unpack(buf)
        remaining_payload = self.payload

        while len(remaining_payload) > 0:
            (_paramid, datasetlength) = struct.unpack(
                IENAQ._FORMAT_, remaining_payload[:IENAQ._FORMAT_LEN_])
            # Check if we have enough payload
            if len(self.payload[IENAQ._FORMAT_LEN_:]) < datasetlength:
                raise Exception("Q Param dataset length {} larger than payload{}".format(
                    datasetlength, len(remaining_payload[IENAQ._FORMAT_LEN_:])))

            qparam = QParameter(paramid=_paramid,
                                dataset=remaining_payload[IENAQ._FORMAT_LEN_:IENAQ._FORMAT_LEN_ + datasetlength])
            self.parameters.append(qparam)
            if datasetlength % 2 == 1:
                padding = 1
            else:
                padding = 0
            remaining_payload = remaining_payload[IENAQ._FORMAT_LEN_ + datasetlength + padding:]

    def pack(self):
        """
        Pack the IENA-Q payload into a binary string

        :rtype: bytes
        """
        self.payload = b""
        for qparam in self:
            datasetlength = len(qparam.dataset)
            self.payload += struct.pack(IENAQ._FORMAT_, qparam.paramid, datasetlength)
            self.payload += qparam.dataset
            if datasetlength % 2 == 1:
                self.payload += struct.pack(">B", 0)

        return super(IENAQ, self).pack()

    def __repr__(self):
        txt = "IENAQ: KEY={:#0X} SEQ={} TIMEUS={} NUM_QPARAM={}\n".format(self.key, self.sequence, self.timeusec, len(self))
        for idx, param in enumerate(self):
            txt += " Q-Param #{}:{}\n".format(idx, repr(param))

        return txt

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.parameters):
            _param = self.parameters[self._index]
            self._index += 1
            return _param
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.parameters)

    def __getitem__(self, key):
        return self.parameters[key]

class DParameter(namedtuple("DParameter", "paramid, delay, dwords")):
    """
    The DParameter is a object representing each DParameter in an IENA-D packet

    :param paramid: The param ID for this parameter
    :type paramid: int
    :param delay: The delay for this parameter
    :type delay: int
    :param dwords: List of all the d-words in the parameter
    :type dwords: list[int]
    """


class IENAD(IENA):
    """
    Support for IENA-D packets. Std parameters with delay field

    This payload includes a defined parameter ID, a delay and a length field

    All this is encapsulated inside an IENA packet as defined above with this pattern repeated
    
    The number of D-words are defined in the N2 field of the IENA header::

        ---2B--- ---2B-- ---2B--- ---2B---- ---2B--- ---2B---
        |PARAM  | DELAY | D#N    | D#-1    | ...    | D0     |
        -------- ------- ------------------------------------  

    Unpack some received packet from the network

    >>> import socket
    >>> recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    >>> data, addr = recv_socket.recvfrom(2048)
    >>> i = IENAD()
    >>> i.unpack(data)
    >>> print len(i.parameters)
    6

    :type parameters: list[DParameter]
    """

    _FORMAT_ = ">HHH"
    REQ_ATTR = ("key", "timeusec", "keystatus", "status", "sequence", "endfield", "payload", "parameters")

    def __init__(self):
        IENA.__init__(self)
        self.parameters = []  #: The list of all DParameters in thie IENA-D packet Each entry is of class :class:`DParameter`
        self._req_attr = IENAD.REQ_ATTR

    def unpack(self, buf):
        """
        Unpack a raw byte stream to an IENA-M object
        Accepts a buffer to unpack as the required argument

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """

        super(IENAD, self).unpack(buf)

        #According to the IENA spec,the N2 field contains the numner of data words
        dataword_count = self.keystatus & 0x7
        len_param_bytes = dataword_count * 2 + 4 # how big is each parameter grouping
        num_params = len(self.payload) // len_param_bytes  # So how many in the payload
        if len(self.payload) - (num_params * len_param_bytes) != 0:
            raise ValueError("There are not an integer number of D Parameters in the payload. Length DParam={} "
                            "Length IENA Payload={}".format(len_param_bytes, len(self.payload)))

        # Loop through
        for param in range(num_params):
            try:
                _dparams = struct.unpack_from(">{}H".format(dataword_count+2), self.payload, param*len_param_bytes)
            except Exception as e:
                raise IndexError("Could not unpack the payload from offset {}. Is the N2 field correct? "
                                "Expected {} data words. Error={}".format(dataword_count, param*len_param_bytes, e))
            dparam = DParameter(paramid=_dparams[0], delay=_dparams[1], dwords=list(_dparams[2:]))
            self.parameters.append(dparam)

    def __repr__(self):
        return "IENAD: KEY={:#0X} SEQ={} TIMEUS={} NUM_DPARAM={}".format(
            self.key, self.sequence, self.timeusec, len(self.parameters))

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.parameters):
            _param = self.parameters[self._index]
            self._index += 1
            return _param
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.parameters)

    def __getitem__(self, key):
        return self.parameters[key]


class NParameter(namedtuple("NParameter", "paramid, dwords")):
    """
    The NParameter is a object representing each NParameter in an IENA-N packet
    
    :param paramid: The param ID for this parameter
    :type paramid: int
    :param dwords: List of all the d-words in the parameter
    :type dwords: list[int]
    """


class IENAN(IENA):
    """
    Support for IENA-N packets. Std parameters without delay field

    This payload includes a defined parameter ID and payload

    All this is encapsulated inside an IENA packet as defined above with this pattern repeated
    
    The number of D-words are defined in the N2 field of the IENA header::

        ---2B--- ---2B-- ---2B--- ---2B---- ---2B--- ---2B---
        |PARAM  | D#N   | D#N-1  |   D#-1  | ...    | D0     |  
        -------- ------- ------------------------------------  

    Unpack some received packet from the network

    >>> import socket
    >>> recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    >>> data, addr = recv_socket.recvfrom(2048)
    >>> i = IENAN()
    >>> i.unpack(data)
    >>> print len(i.parameters)
    6
    >>> print i.parameters[0].paramid
    >>>
    4

    :type parameters: list[NParameter]
    
    """

    _FORMAT_ = ">HHH"
    REQ_ATTR = ("key", "timeusec", "keystatus", "status", "sequence", "endfield", "payload", "parameters")

    def __init__(self):
        IENA.__init__(self)
        self.parameters = []  #: List of all N-type parameters. Each entry is of class :class:`NParameter`
        self._req_attr = IENAN.REQ_ATTR

    def unpack(self, buf):
        """
        Unpack a raw byte stream to an IENA-M object
        Accepts a buffer to unpack as the required argument

        :param buf: The string buffer to unpack
        :type buf: bytes

        :rtype: bool
        """

        super(IENAN, self).unpack(buf)

        #According to the IENA spec,the N2 field contains the numner of data words
        dataword_count = self.keystatus & 0x7
        len_param_bytes = dataword_count * 2 + 2 # how big is each parameter grouping
        num_params = len(self.payload) // len_param_bytes  # So how many in the
        # Check that we don't have od
        if len(self.payload) - (num_params * len_param_bytes) != 0:
            raise ValueError("There are not an integer number of N Parameters in the payload. Length NParam={} "
                            "Length IENA Payload={}".format(len_param_bytes, len(self.payload)))

        # Loop through
        for param in range(num_params):
            try:
                _nparams = struct.unpack_from(">{}H".format(dataword_count+1), self.payload, param*len_param_bytes)
            except Exception as e:
                raise IndexError("Could not unpack the payload from offset {}. Is the N2 field correct? "
                                "Expected {} data words. Error={}".format(dataword_count, param*len_param_bytes, e))
            nparam = NParameter(paramid=_nparams[0], dwords=list(_nparams[1:]))
            self.parameters.append(nparam)

    def __repr__(self):
        return "IENAN: KEY={:#0X} SEQ={} TIMEUS={} NUM_DPARAM={}".format(
            self.key, self.sequence, self.timeusec, len(self))

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.parameters):
            _param = self.parameters[self._index]
            self._index += 1
            return _param
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.parameters)

    def __getitem__(self, key):
        return self.parameters[key]



"""
.. module:: ParserAlignedPacket
    :platform: Unix, Windows
    :synopsis: Class to pack and unpack Parser Aligned payloads

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

import struct


__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


class ParserAlignedPacket(object):
    """
    A class that handles parser aligned packets. Unpack a buffer to populate the field into a list of parserblocks
    
    Capture a UDP packet,unpack as iNetX whose payload is parser aligned
    
    >>> import AcraNetwork.iNetX as inetx
    >>> import socket
    >>> recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    >>> data, addr = recv_socket.recvfrom(2048)
    >>> i = inetx.iNetX()
    >>> i.unpack(data)
    >>> p = ParserAlignedPacket()
    >>> p.unpack(i.payload)
    >>> print p
    
    :type parserblocks: list[ParserAlignedBlock]
    :type numberofblocks: int
    """

    def __init__(self):
        self.parserblocks = [] #: List of ParserAlignedBlock
        self.numberofblocks = 0 #: The number of ParserAlignedBlock

    def unpack(self, buf):
        """
        Pass a buffer containing all the Parser alignd payload and this method will
        unpack all the fields. Returns a list of parserblocks
    
        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """
        fullbufferlen = len(buf)
        bufferparsed = 0
        while bufferparsed < fullbufferlen:
            block = ParserAlignedBlock()
            # unpack and add the amount unpacked to the running total
            bufferparsed += block.unpack(buf[bufferparsed:])
            self.parserblocks.append(block)
        return True

    def pack(self):
        """Convert a ParserPacket to a buffer"""

        buf = bytes()
        for b in self.parserblocks:
            buf += b.pack()

        return buf

    def __iter__(self):
        self._index = 0
        return self

    def next(self):
        if self._index < len(self.parserblocks):
            _block = self.parserblocks[self._index]
            self._index += 1
            return _block
        else:
            raise StopIteration

    __next__ = next

    def __len__(self):
        return len(self.parserblocks)

    def __getitem__(self, key):
        return self.parserblocks[key]

    def __repr__(self):
        rep = ""
        for idx, b in enumerate(self.parserblocks):
            rep += "Block {}: {}\n".format(idx, repr(b))
        return rep

    def __eq__(self, other):
        if not isinstance(other, ParserAlignedPacket):
            return False
        if len(other) != len(self):
            return False
        blk_cnt = len(other)
        for idx in range(blk_cnt):
            if other.parserblocks[idx] != self.parserblocks[idx]:
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)

class ParserAlignedBlock(object):
    """
    A class to handle a single Parser Block. Returns an object containing all the fields of a parser block
    
    :type error: bool
    :type errorcode: int
    :type quadbytes: int
    :type messagecount: int
    :type busid: int
    :type elapsedtime: int
    :type payload: str
    """
    def __init__(self):

        self.error = False #: Error Flag
        self.errorcode = 0 #: Error code field
        self.quadbytes = 1 #: Number of quadbytes in the parser block
        self.messagecount = 0 #: Wrapping 8 bit message counter
        self.busid = 1 #: Bus ID on which the message was captured
        self.elapsedtime = 2 #: Time tag in nanoseconds offset from the iNetx timestamp
        self.payload = "" #: Payload

        self.format = '>HBBL'
        self.headerlen = struct.calcsize(self.format)

    def unpack(self, buf):
        """
        Unpack a single parser block. Unsually called only from the ParserAlignedPacket unpack method. 
        
        Returns the length of the parser block
        
        :param buf: The string buffer to unpack
        :type buf: str
        :rtype: int
        """

        # Unpack the fields
        (error_and_quad, self.messagecount, self.busid, self.elapsedtime)= struct.unpack_from(self.format, buf)
        # MSB bit is the error flag
        if (error_and_quad >> 15) == 1:
            self.error = True
        else:
            self.error = False

        # As error code and quad bytes are not byte aligned I have to do some shifting and masking
        self.errorcode = (error_and_quad >> 9) & 0x3f
        self.quadbytes = error_and_quad & 0x1ff
        payload_length_in_bytes = (self.quadbytes - 2)*4

        if self.quadbytes < 2:
            raise ValueError("The quad bytes cannot be less than 2. Actual={}".format(self.quadbytes))

        # Raise an exception if the payload is not as expected
        if len(buf) < self.headerlen + payload_length_in_bytes:
            raise ValueError("The length of the block buffer {} is smaller the expect length {}".format(
                len(buf), self.headerlen + payload_length_in_bytes
            ))

        self.payload = buf[self.headerlen:self.quadbytes*4]
        # Return the length of the block as a courtesy
        return self.quadbytes*4

    def pack(self):
        """
        Convert a ParserAlignedBlock into a buffer
        :return:
        """
        if len(self.payload) % 4 != 0:
            raise Exception("Length of payload is not aligned to quadbytes (32b)")
        self.quadbytes = 2 + len(self.payload) // 4
        error_and_quad = (self.error << 15) + ((self.errorcode & 0x3f) << 9) + self.quadbytes
        return struct.pack(self.format, error_and_quad, self.messagecount, self.busid, self.elapsedtime) + self.payload

    def __repr__(self):
        return "QuadBytes={} Error={} ErrorCode={} BusID={} MessageCount={} ElapsedTime={}".format(
            self.quadbytes, self.error, self.errorcode, self.busid, self.messagecount, self.elapsedtime
        )

    def __len__(self):
        return len(self.payload) + struct.calcsize(self.format)

    def __eq__(self, other):
        if not isinstance(other, ParserAlignedBlock):
            return False
        for attr in ["quadbytes", "error", "errorcode", "busid", "messagecount", "elapsedtime", "payload"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __ne__(self, other):
        return not self.__eq__(other)


class ARINC429(object):
    """This is not working yet. Don't use it"""
    MESSAGE_LEN = 4 # 4 bytes
    LABEL_REVERSE = [
        0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0, #   0 -   8
        0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0, #   9 -  15
        0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8, #  16 -  23
        0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8, #  24 -  31
        0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4, #  32 -  39
        0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4, #  40 -  47
        0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec, #  48 -  55
        0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc, #  56 -  63
        0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2, #  64 -  71
        0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2, #  72 -  79
        0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea, #  80 -  87
        0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa, #  88 -  95
        0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6, #  96 - 103
        0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6, # 104 - 111
        0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee, # 112 - 119
        0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe, # 120 - 127
        0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1, # 128 - 135
        0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1, # 136 - 143
        0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9, # 144 - 151
        0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9, # 152 - 159
        0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5, # 160 - 167
        0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5, # 168 - 175
        0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed, # 176 - 183
        0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd, # 184 - 191
        0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3, # 192 - 199
        0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3, # 200 - 207
        0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb, # 208 - 215
        0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb, # 216 - 223
        0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7, # 224 - 231
        0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7, # 232 - 239
        0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef, # 240 - 247
        0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff, # 248 - 255
        0x00
    ]

    def __init__(self):
        self.parity = None
        self.ssm = None
        self.data = None
        self.sdi = None
        self.label = None

    def unpack(self,buf):
        if len(buf) != ARINC429.MESSAGE_LEN:
            raise ValueError("Buffer is not the correct length for an ARINC429 message")
        (byte1,byte2,byte3,byte4) = struct.unpack('BBBB',buf)
        self.parity = byte1 / 128
        self.ssm = (byte1/32) % 4
        self.data = ((byte1 %32) * 256 + byte2 ) * 64 + (byte3 /4)
        self.sdi = byte3 % 4
        self.label = ARINC429.LABEL_REVERSE[byte4]






"""
.. module:: iNetX
    :platform: Unix, Windows
    :synopsis: Class to construct and de construct iNetx Packets

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct


class iNetX (object):
    """ 
    Class to pack and unpack iNetX payloads. iNet-X is an open payload format for use
    in FTI networks. It is usually transmitted in a UDP packet containing parameter data
    acquired from sensors and buses
    
    Capture a UDP packet and unpack the payload as an iNetX packet
    
    >>> import socket
    >>>> recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  
    >>> data, addr = recv_socket.recvfrom(2048)
    >>> i = iNetX()
    >>> i.unpack(data)
    >>> print i.streamid
    6
    
    :type inetxcontrol: int
    :type streamid: int
    :type sequence: int
    :type packetlen: int
    :type ptptimeseconds: int
    :type ptptimenanoseconds: int
    :type pif: int
    :type payload: bytes
    """

    DEF_CONTROL_WORD = 0x11000000 #:(Object Constant) The default iNetX control word.
    INETX_HEADER_FORMAT = '>LLLLLLL'
    INETX_HEADER_LENGTH = struct.calcsize(INETX_HEADER_FORMAT)
    REQ_ATTR = ("inetxcontrol", "streamid", "sequence", "ptptimeseconds", "ptptimenanoseconds", "pif", "payload")

    def __init__(self, buf=None):
        '''Creator method for an iNetX class'''
        self.inetxcontrol = iNetX.DEF_CONTROL_WORD #: Control Word
        self.streamid = None #: Stream ID. Typically to identify a unique packet in an FTI network. 4 bytes in size
        self.sequence = None #: Unique rollover counter per stream ID.Rolls over at 2^64
        self.packetlen = None #: Packet Length
        self.ptptimeseconds = None #: Timestamp of first parameter in the packet. EPOCH time
        self.ptptimenanoseconds = None #: Nanaosecond timestamp
        self.pif = None #: Payload Information Field
        self.payload = None #: Payload

        self._packetStrut = struct.Struct(iNetX.INETX_HEADER_FORMAT)
        if buf is not None:
            self.unpack(buf)

    def pack(self):
        """
        Pack the packet into a binary format and return as a string
        
        :rtype: bytes
        """

        for attr in iNetX.REQ_ATTR:
            if getattr(self, attr) is None:
                raise ValueError("Require {} is not defined".format(attr))

        self.packetlen =  len(self.payload)  + iNetX.INETX_HEADER_LENGTH
        packetvalues = (self.inetxcontrol,self.streamid,self.sequence,self.packetlen,self.ptptimeseconds,self.ptptimenanoseconds,self.pif )
        packet = self._packetStrut.pack(*packetvalues) + self.payload
        return packet

    def unpack(self, buf):
        """
        Unpack a raw byte stream to an iNetX object
        Accepts a buffer to unpack as the required argument

        :param buf: The string buffer to unpack
        :type buf: bytes
        :rtype: bool
        """

        if len(buf) < iNetX.INETX_HEADER_LENGTH:
            raise ValueError ("Buffer is too short to be an iNetX packet")

        self.inetxcontrol,self.streamid,self.sequence,self.packetlen,self.ptptimeseconds,self.ptptimenanoseconds,self.pif  = self._packetStrut.unpack_from(buf)

        if self.packetlen != len(buf):
            raise ValueError("Length of buffer 0x{:X} does not match length field 0x{:X}".format(len(buf),self.packetlen))

        self.payload = buf[iNetX.INETX_HEADER_LENGTH:]

        return True

    def setPacketTime(self,utctimestamp, nanoseconds=0):
        """
        Set the packet timestamp
        
        :param timestamp: The timestamp in seconds since 1 Jan 1970
        :type timestamp: int
        :type nanoseconds: Nanoseconds past the current time
        :type nanoseconds: int
        """

        self.ptptimeseconds = utctimestamp
        self.ptptimenanoseconds = nanoseconds

        return True

    def __repr__(self):
        return "STREAMID={:#0X} SEQ={} LEN={} PTPS={} PTPNS={}".format(self.streamid, self.sequence, self.packetlen,
                                                                       self.ptptimeseconds, self.ptptimenanoseconds)

    def __eq__(self, other):
        if not isinstance(other, iNetX):
            return False

        for attr in iNetX.REQ_ATTR:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __len__(self):
        return len(self.pack())





import unittest
__author__ = 'diarmuid'
import sys
sys.path.append("..")
import unittest
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.Chapter10 as ch10
import struct
import os
import random
import copy
import datetime


THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def getEthernetPacket(data=""):
    e = SimpleEthernet.Ethernet()
    e.srcmac = 0x001122334455
    e.dstmac = 0x998877665544
    e.type = SimpleEthernet.Ethernet.TYPE_IP
    i = SimpleEthernet.IP()
    i.dstip = "235.0.0.1"
    i.srcip = "192.168.1.1"
    i.protocol = SimpleEthernet.IP.PROTOCOLS["UDP"]
    u = SimpleEthernet.UDP()
    u.dstport = 6679
    u.srcport = 6679
    u.payload = data
    i.payload = u.pack()
    e.payload = i.pack()
    return e.pack()


def wrap_in_udp_and_pcap(mybuffer, pcapf, mode="w"):
    pcapw = pcap.Pcap(pcapf, mode=mode)
    pcapw.write_global_header()
    rec = pcap.PcapRecord()
    rec.payload = getEthernetPacket(mybuffer)
    pcapw.write(rec)
    pcapw.close()
    return True

arinc_packet="""ARINCPayload: MessageCount=11
  ARINCData: GapTime=0 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=4000 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3900 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3600 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=4100 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=4000 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3700 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3600 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=4000 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3600 FormatError=False ParityError=False BusSpeed=0 Bus=23
  ARINCData: GapTime=3900 FormatError=False ParityError=False BusSpeed=0 Bus=23
"""



def get_ch10(len=4):
    c = ch10.Chapter10()
    c.channelID = 1
    c.datatypeversion = 2
    c.sequence = 3
    c.packetflag = 0  # No secondary
    c.datatype = 4
    c.relativetimecounter = 100
    c.payload = os.urandom(len)
    return c


class CH10UDPTest(unittest.TestCase):
    def setUp(self):

        self.full = ch10.Chapter10UDP()
        self.full.type = ch10.Chapter10UDP.TYPE_FULL
        self.full.sequence = 1
        self.full.chapter10.channelID = 1
        self.full.chapter10.datatypeversion = 2
        self.full.chapter10.sequence = 3
        self.full.chapter10.packetflag = 0 # No secondary
        self.full.chapter10.datatype = 4
        self.full.chapter10.relativetimecounter = 100

        self.seg = ch10.Chapter10UDP()
        self.seg.sequence = 2
        self.seg.type = ch10.Chapter10UDP.TYPE_SEG
        self.seg.channelID = 0x232
        self.seg.channelsequence = 100
        self.seg.segmentoffset = 3
        self.seg.chapter10.channelID = 1
        self.seg.chapter10.datatypeversion = 2
        self.seg.chapter10.sequence = 3
        self.seg.chapter10.packetflag = 0 # No secondary
        self.seg.chapter10.datatype = 4
        self.seg.chapter10.relativetimecounter = 100

        self.c = ch10.Chapter10()
        self.c.channelID = 1
        self.c.datatypeversion = 2
        self.c.sequence = 3
        self.c.packetflag = 0 # No secondary
        self.c.datatype = 4
        self.c.relativetimecounter = 100

    def test_ch10_to_pcap(self):
        #self.full._payload = struct.pack(">II", 33, 44)
        full_payload = getEthernetPacket(self.full.pack())
        self.pcapw = pcap.Pcap("test_ch10.pcap", mode="w")
        self.pcapw.write_global_header()
        self.rec = pcap.PcapRecord()
        self.rec.payload = full_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        #self.seg._payload = struct.pack(">II", 55, 66)
        seg_payload = getEthernetPacket(self.seg.pack())
        self.rec.payload = seg_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        self.pcapw.close()

    def test_ch10_eq(self):
        full = ch10.Chapter10UDP()
        full.type = ch10.Chapter10UDP.TYPE_FULL
        full.sequence = 1
        #full._payload = struct.pack(">II", 33, 44)

        #self.full._payload = struct.pack(">II", 33, 44)

        self.assertTrue(full==self.full)

    def test_unpack(self):
        #self.full._payload = struct.pack(">II", 33, 44)
        #self.seg._payload = struct.pack(">II", 33, 44)

        full_unpack = ch10.Chapter10UDP()
        self.assertTrue(full_unpack.unpack(self.full.pack()))
        #print(len(self.full.payload))
        #print(len(full_unpack.payload))

        self.assertTrue(full_unpack == self.full)
        self.assertFalse(full_unpack == self.seg)

    def test_ch10_eqseg(self):
        seg = ch10.Chapter10UDP()
        seg.type = ch10.Chapter10UDP.TYPE_SEG
        seg.sequence = 2
        seg.channelID = 0x232
        seg.channelsequence = 100
        seg.segmentoffset = 3
        #seg._payload = struct.pack(">II", 33, 44)

        #self.seg._payload = struct.pack(">II", 33, 44)

        self.assertTrue(seg==self.seg)

    def test_ch10pay_to_pcap(self):
        self.c.payload = struct.pack(">II", 33, 44)
        #self.full._payload = self.c.pack()
        full_payload = getEthernetPacket(self.full.pack())
        self.pcapw = pcap.Pcap("test_ch10_2.pcap", mode="w")
        self.pcapw.write_global_header()
        self.rec = pcap.PcapRecord()
        self.rec.payload = full_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        self.pcapw.close()

    def test_ch10_comp(self):
        ref = ch10.Chapter10()
        #self.c._payload = struct.pack(">II", 33, 44)
        pay = self.c.pack()
        ref.unpack(pay)
        self.assertTrue(ref == self.c)

    def test_ch10_sec_hdr(self):
        c = ch10.Chapter10UDP()
        c.type = ch10.Chapter10UDP.TYPE_FULL
        c.sequence = 1
        c.chapter10.channelID = 1
        c.chapter10.datatypeversion = 2
        c.chapter10.sequence = 3
        c.chapter10.packetflag = ch10.Chapter10.PKT_FLAG_SECONDARY
        c.chapter10.datatype = 4
        c.chapter10.ts_source = "ieee1588"
        c.chapter10.ptptimeseconds = 101
        c.chapter10.ptptimenanoseconds = int(200e6)
        c.chapter10.relativetimecounter = 0x0
        c.chapter10.payload = struct.pack(">QQ", 33, 44)
        full_payload = getEthernetPacket(c.pack())
        self.pcapw = pcap.Pcap("test_ch10_3.pcap", mode="w")
        self.pcapw.write_global_header()
        self.rec = pcap.PcapRecord()
        self.rec.payload = full_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        self.pcapw.close()

    def test_ch10_format2(self):
        self.pcapw = pcap.Pcap("test_ch10_format2.pcap", mode="w")
        self.pcapw.write_global_header()
        self.rec = pcap.PcapRecord()
        c = ch10.Chapter10UDP()
        c.format = 2
        c.type = ch10.Chapter10UDP.TYPE_FULL
        c.sequence = 10
        c.segmentoffset = 5000
        c.channelID = 20
        c.chapter10 = get_ch10(16)
        self.assertTrue(c.pack())

        full_payload = getEthernetPacket(c.pack())
        self.rec.payload = full_payload
        self.assertIsNone(self.pcapw.write(self.rec))
        self.pcapw.close()

        c2 = ch10.Chapter10UDP()
        c2.unpack(c.pack())
        self.assertTrue(c2==c)


    def test_ch10_format3(self):
        pcapw = pcap.Pcap("test_ch10_format3.pcap", mode="w")
        pcapw.write_global_header()
        rec = pcap.PcapRecord()
        for len in range(5):
            c = ch10.Chapter10UDP()

            c.format = 3
            c.sourceid_len = len
            if len > 0:
                c.sourceid = 0x4
            c.offset_pkt_start = 4
            c.sequence = 0x15
            c.chapter10 = get_ch10(16)
            self.assertTrue(c.pack())
            c2 = ch10.Chapter10UDP()
            c2.unpack(c.pack())
            self.assertTrue(c2 == c)
            for packet in [c, c2]:
                full_payload = getEthernetPacket(packet.pack())
                rec.payload = full_payload
                self.assertIsNone(pcapw.write(rec))
        pcapw.close()

    def test_arinc_payload(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_arinc.pcap"))
        mypcaprecord = p[0]
        p.close()
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        ip = SimpleEthernet.IP()
        ip.unpack(e.payload)
        u = SimpleEthernet.UDP()
        u.unpack(ip.payload)
        # Now I have a _payload that will be an inetx packet
        c = ch10.Chapter10UDP()
        self.assertTrue(c.unpack(u.payload))

        # Check the Ch10 packets
        self.assertEqual(c.chapter10.syncpattern, 0xeb25)
        self.assertEqual(c.chapter10.channelID, 0x000080d6)
        self.assertEqual(c.chapter10.datatypeversion, 0x44)
        # nanoseconds and seconds
        self.assertEqual(c.chapter10.ptptimenanoseconds, 1237463)
        self.assertEqual(c.chapter10.ptptimeseconds, 0)

        self.assertEqual(repr(c), "CH10 UDP Full Packet: Format=1 Sequence=0 Payload=Chapter 10: ChannelID=32982 Sequence=0 DataLen=88")
        arinc_p = ch10.ARINC429DataPacket()
        self.assertTrue(arinc_p.unpack(c.chapter10.payload))
        self.assertEqual(len(arinc_p.arincwords), 11)
        self.assertEqual(arinc_p.arincwords[0].payload, struct.pack(">I", 0x4bd91e2e))

        arinc2 = ch10.ARINC429DataPacket()
        self.assertTrue(arinc2.unpack(arinc_p.pack()))
        self.assertTrue(arinc_p == arinc2)

        self.assertEqual(repr(arinc2), arinc_packet)
        for idx, aw in enumerate(arinc2):
            if idx == 2:
                self.assertEqual(repr(aw), "ARINCData: GapTime=3900 FormatError=False ParityError=False BusSpeed=0 Bus=23")

    def test_format1_looks_like_2(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_format_1_looks_like2.pcap"))
        for r in p:
            c = ch10.Chapter10UDP()
            c.unpack(r.payload[0x2a:])
            self.assertEqual(c.format, 1)
        p.close()


uart_pkt="""UARTPayload: UARTDataWordCount=1
  UARTDataWord: PTPSec=201450103 PTPNSec=10 ParityError=False DataLen=508 SubChannel=8191
"""
class Ch10UARTTest(unittest.TestCase):
    def setUp(self):

        self.full = ch10.Chapter10UDP()
        self.full.type = ch10.Chapter10UDP.TYPE_FULL
        self.full.sequence = 10
        self.full.chapter10.channelID = 23
        self.full.chapter10.datatypeversion = 2
        self.full.chapter10.sequence = 3
        self.full.chapter10.packetflag = 0xC4  # Secondary time + PTP
        self.full.chapter10.datatype = 0x50
        self.full.chapter10.relativetimecounter = 100
        self.full.chapter10.ptptimeseconds = 22
        self.full.chapter10.ptptimenanoseconds = 250000

    def test_basic_uart(self):
        udp = ch10.UARTDataPacket()
        for i in range(5):
            udw = ch10.UARTDataWord()
            udw.ptptimeseconds = 22*i
            udw.ptptimenanoseconds = 250000+i
            udw.subchannel = 52 + 2*i
            udw.parity_error = random.choice([True, False])
            udw.payload = os.urandom(random.randint(15, 220))
            #udw.payload = os.urandom(16)

            self.assertIsNone(udp.append(udw))

        self.full.chapter10.payload = udp.pack()
        self.assertTrue(wrap_in_udp_and_pcap(self.full.pack(), "ch10_uart.pcap"))

        # Test unpack and comparsion
        dummy_ch10 = ch10.Chapter10UDP()
        dummy_ch10.unpack(self.full.pack())
        self.assertTrue(dummy_ch10 == self.full)
        self.assertEqual(repr(udp[0]), "UARTDataWord: PTPSec=0 PTPNSec=250000 ParityError={} DataLen={} "
                                       "SubChannel=52".format(udp[0].parity_error, udp[0].datalength))

    def test_uart_unpack(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_uart2.pcap"))
        mypcaprecord = p[0]
        ch10pkt = ch10.Chapter10UDP()
        ch10pkt.unpack(mypcaprecord.payload[0x2a:-4]) # FCS
        self.assertEqual(ch10pkt.chapter10.packetflag, 0xc4)
        uart = ch10.UARTDataPacket()
        uart.unpack(ch10pkt.chapter10.payload)
        self.assertEqual(len(uart), 1)
        (b1,b2) = struct.unpack_from("<BB", uart[0].payload)
        self.assertEqual(b1, 0xd5)
        p.close()
        for idx, uartdw in enumerate(uart):
            if idx == 0:
                dw2 = copy.copy(uartdw)
                self.assertEqual(dw2, uartdw)
                #print("{}:{}:{}".format(idx,s_idx,0))

        uart2 = copy.copy(uart)
        self.assertEqual(uart, uart2)
        self.assertEqual(repr(uart2),uart_pkt)

    @unittest.skip("")
    def test_fmt2_unpack(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_fmt_1.pcap"))
        mypcaprecord = p[0]
        ch10pkt = ch10.Chapter10UDP()
        ch10pkt.unpack(mypcaprecord.payload[0x2a:-4])  # FCS
        self.assertEqual(ch10pkt.format, 1)
        #print(repr(ch10pkt.chapter10))


class Time_Test(unittest.TestCase):


    def test_playground(self):
        ms = 989
        self.assertEqual(0x98, ch10.double_digits_to_bcd(ms/10))
        self.assertEqual(0x98, ch10.double_digits_to_bcd(ms/10))

        #05/20/19 16:41:24'
        current_time = datetime.datetime.fromtimestamp(1558366884)
        self.assertEqual(0x41, ch10.double_digits_to_bcd(current_time.minute))
        self.assertEqual(0x24, ch10.double_digits_to_bcd(current_time.second))
        #self.assertEqual(0x16, ch10.double_digits_to_bcd(current_time.hour))

        self.assertEqual(99, ch10.bcd_to_int(0x99))
        self.assertEqual(9999, ch10.bcd_to_int(0x9999))

    def test_time_pkt(self):
        t = ch10.TimeDataFormat1()
        t.milliseconds = 980
        t.datetime = datetime.datetime.fromtimestamp(1558366884)

        #print(repr(t))
        t2 = ch10.TimeDataFormat1()
        t2.unpack(t.pack())
        self.assertTrue(t == t2)
        #print(repr(t))
        #print(repr(t2))

    def test_time_pkt_2(self):
        t = ch10.TimeDataFormat2()
        t.nanoseconds = 999999999
        t.datetime = datetime.datetime.fromtimestamp(1558366884)

        t2 = ch10.TimeDataFormat2()
        t2.unpack(t.pack())
        self.assertTrue(t2 == t)
        #print(repr(t))
        #self.assertEqual("TimeFormat2 ChannelSpecificWord=0X11 Time=16:41:24 05/20/19 20-May 2019 NanoSeconds=999999999", repr(t))
        #print(repr(t2))

    def test_time_to_pcap(self):
        pcapw = pcap.Pcap("test_ch10_time.pcap", mode="w")
        pcapw.write_global_header()
        rec = pcap.PcapRecord()
        cu = ch10.Chapter10UDP()
        cu.type = ch10.Chapter10UDP.TYPE_FULL
        cu.sequence = 1
        c = get_ch10()
        types = [ch10.DATA_TYPE_TIMEFMT_1, ch10.DATA_TYPE_TIMEFMT_2]
        t1 = ch10.TimeDataFormat1()
        t2 = ch10.TimeDataFormat2()
        for i,t in enumerate([t1, t2]):
            t.milliseconds = 980
            t.datetime = datetime.datetime.fromtimestamp(1558366884)
            c.payload = t.pack()
            c.datatype = types[i]
            cu.chapter10 = c
            full_payload = getEthernetPacket(cu.pack())
            rec.payload = full_payload
            self.assertIsNone(pcapw.write(rec))
        pcapw.close()

    @unittest.skip("")
    def test_ch10_fmt_pcap(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ch10_pkt.pcap"))
        mypcaprecord = p.next()
        c = ch10.Chapter10UDP()
        c.unpack(mypcaprecord.payload[0x2a:])
        t = ch10.TimeDataFormat1()
        t.unpack(c.chapter10.payload)
        #print(t.datetime.strftime("%c"))
        mypcaprecord = p.next()
        c = ch10.Chapter10UDP()
        c.unpack(mypcaprecord.payload[0x2a:])
        t = ch10.TimeDataFormat2()
        t.unpack(c.chapter10.payload)
        #print(t.datetime.strftime("%c"))
        self.assertEqual(t.datetime, datetime.datetime.fromtimestamp(1561535306))



class GenPCAP(unittest.TestCase):

    def test_ch10_fmt1(self):
        pcapw = pcap.Pcap("test_ch10_sample.pcap", mode="w")
        pcapw.write_global_header()
        rec = pcap.PcapRecord()
        for fmt in [1, 2, 3]:
            c = get_ch10(16)
            c.ptptimeseconds = 333344
            c.ptptimenanoseconds = 42322
            c.packetflag = ch10.Chapter10.PKT_FLAG_SECONDARY + ch10.Chapter10.PKT_FLAG_SEC_HDR_TIME + ch10.Chapter10.PKT_FLAG_1588_TIME
            cu = ch10.Chapter10UDP()
            cu.type = ch10.Chapter10UDP.TYPE_FULL
            cu.sequence = 1
            cu.chapter10 = c
            # format 2
            cu.segmentoffset = 0x0
            cu.channelID = c.channelID
            # Format 3
            cu.sourceid_len = 0x1
            cu.sourceid = 0x2
            cu.sequence = 0x3
            cu.offset_pkt_start = 0x0

            cu.format = fmt

            # Write it out

            full_payload = getEthernetPacket(cu.pack())
            rec.payload = full_payload
            self.assertIsNone(pcapw.write(rec))
        pcapw.close()


class Ch10Mil(unittest.TestCase):

    def test_mil(self):

        m = ch10.MILSTD1553DataPacket()
        msg = ch10.MILSTD1553Message()
        msg.message = struct.pack(">II", 1,2)
        msg.ptptimeseconds = 100
        msg.ptptimenanoseconds = 200
        m.append(msg)
        m2 = ch10.MILSTD1553DataPacket()
        m2.unpack(m.pack())
        self.assertEqual(m2, m)


class PCMData(unittest.TestCase):

    def test_pcm(self):
        payload_len = 10
        pcapw = pcap.Pcap("test_ch10_pcm.pcap", mode="w")
        rec = pcap.PcapRecord()
        u = ch10.Chapter10UDP()
        u.type = ch10.Chapter10UDP.TYPE_FULL
        u.sequence = 0
        u.chapter10 = get_ch10()
        u.chapter10.datatype = 0x9
        pcmdf = ch10.PCMDataPacket()
        pcmdf.channel_specific_word = 0x1234
        for mfcount in range(1, 4):
            mf = ch10.PCMMinorFrame()
            mf.intra_packet_data_header = mfcount
            mf.intra_packet_sec = 2 * mfcount
            mf.intra_packet_nsec = 4 * mfcount
            mf.minor_frame_data = os.urandom(payload_len)
            pcmdf.minor_frames.append(mf)
        packed_data = pcmdf.pack()
        u.chapter10.payload = packed_data
        rec.set_current_time()
        rec.payload = getEthernetPacket(u.pack())
        pcapw.write(rec)
        pcapw.close()

        pcmdf2 = ch10.PCMDataPacket()
        pcmdf2.minor_frame_size_bytes = payload_len
        pcmdf2.unpack(packed_data)
        self.assertEqual(pcmdf, pcmdf2)

if __name__ == '__main__':
    unittest.main()



__author__ = 'diarmuid'
import sys
sys.path.append("..")
import os

import unittest
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet
import struct


THIS_DIR = os.path.dirname(os.path.abspath(__file__))


def getEthernetPacket(data=0xa):
    e = SimpleEthernet.Ethernet()
    e.srcmac = 0x001122334455
    e.dstmac = 0x998877665544
    e.type = SimpleEthernet.Ethernet.TYPE_IP
    e.payload = struct.pack("H",data)
    return e.pack()


class PcapBasicTest(unittest.TestCase):

    def test_missingfilename(self):
        self.assertRaises(TypeError,lambda: pcap.Pcap())

    def test_missingreadfile(self):
        self.assertRaises(IOError, lambda: pcap.Pcap(os.path.join(THIS_DIR, "nofile.pcap")))

    def test_defaultMagicNumber(self):
        p = pcap.Pcap("_tmp.pcap", mode='w')
        self.assertEqual(p.magic, 0xa1b2c3d4)
        p.close()

    def test_defaultVersionMaj(self):
        p = pcap.Pcap("_tmp.pcap", mode='w')
        self.assertEqual(p.versionmaj, 2)
        p.close()

    def test_defaultVersionMin(self):
        p = pcap.Pcap("_tmp.pcap", mode='w')
        self.assertEqual(p.versionmin, 4)
        p.close()

    def test_readTestFile(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "test_input.pcap"))
        self.assertEqual(p.magic,0xa1b2c3d4)
        self.assertEqual(p.network,1)
        self.assertEqual(p.sigfigs,0)
        self.assertEqual(p.snaplen,262144)
        self.assertEqual(p.versionmaj,2)
        self.assertEqual(p.versionmin,4)
        self.assertEqual(p.zone,0)
        self.assertEqual(p.filesize, 704)
        for idx, rec in enumerate(p):
            if id == 3:
                self.assertEqual(len(rec), 66)
        p.close()



    def test_readARecord(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "test_input.pcap"))
        mypcaprecord = p[0]
        p.close()
        self.assertEqual(mypcaprecord.sec,1419678111)
        self.assertEqual(mypcaprecord.usec,811463)
        self.assertEqual(mypcaprecord.orig_len,70)
        self.assertEqual(mypcaprecord.incl_len,70)
        self.assertEqual(repr(mypcaprecord), "LEN:70 SEC:1419678111 USEC:811463")
        p.close()

    def test_writeARecord(self):
        p = pcap.Pcap("_tmp.pcap",mode='w')
        p.write_global_header()
        r = pcap.PcapRecord()
        r.setCurrentTime()
        r.packet = getEthernetPacket(0xa)
        p.write(r)
        p.close()
        p = pcap.Pcap("_tmp.pcap")
        self.assertEqual(p.magic,0xa1b2c3d4)
        self.assertEqual(p.network,1)
        self.assertEqual(p.sigfigs,0)
        self.assertEqual(p.snaplen,65535)
        self.assertEqual(p.versionmaj,2)
        self.assertEqual(p.versionmin,4)
        self.assertEqual(p.zone,0)
        self.assertEqual(p.filesize,56)
        p.close()
        os.remove("_tmp.pcap")

    def test_appendARecord(self):
        p = pcap.Pcap("_tmp2.pcap",mode='w')
        p.write_global_header()
        r = pcap.PcapRecord()
        r.setCurrentTime()
        r.packet = getEthernetPacket(0xa)
        p.write(r)
        p.close()
        # Now try to append a record
        p = pcap.Pcap("_tmp2.pcap",mode='a')
        r.packet = getEthernetPacket(0xb)
        p.write(r)
        p.close()
        # Read back to verify
        p = pcap.Pcap("_tmp2.pcap")
        self.assertEqual(p.filesize,88)
        for idx, rec in enumerate(p):
            e = SimpleEthernet.Ethernet()
            e.unpack(rec.packet)
            if idx == 0:
                self.assertEqual(e.payload,struct.pack("H",0xa))
            else:
                self.assertEqual(e.payload,struct.pack("H",0xb))
        p.close()
        os.remove("_tmp2.pcap")


if __name__ == '__main__':
    unittest.main()





"""
.. module:: pcap
    :platform: Unix, Windows
    :synopsis: Class to pack and unpack pcap files

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""

__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
import os
import time
import warnings


class PcapRecord(object):
    """
    Class that can be used to store one pcap record. A Pcap file containts one or more PcapRecords
    
    :type sec: int
    :type usec: int
    :type incl_len: int
    :type orig_len: int
    :type _packet: str
    """
    def __init__(self):
        self.sec = 0 #: Second timestamp of the record. Epoch time
        self.usec = 0 #: Microsecond timestamp of the record
        self.incl_len = None  #: The number of bytes captured and saved in the file
        self.orig_len = None  #: The number of bytesas appearded on the network when captured
        self._packet = None

    # Use a property on packet so that the length is triggered on it changing
    @property
    def packet(self):
        """
        The payload within the pcap record. Payload is more accurate

        :rtype: bytes
        """
        return self._packet

    @packet.setter
    def packet(self,p):
        self._packet = p
        self.incl_len = len(p)
        self.orig_len = self.incl_len

    @property
    def payload(self):
        """
        The payload within the pcap record.
        
        :rtype: bytes
        """
        return self._packet

    @payload.setter
    def payload(self,p):
        self._packet = p
        self.incl_len = len(p)
        self.orig_len = self.incl_len

    def unpack(self, buf):
        """
        Unpack the pcap header. Pass in a buffer containing the header
        
        :type buf: bytes
        """

        if struct.calcsize(Pcap.RECORD_HEADER_FORMAT) != len(buf):
            raise ValueError("Header buffer is not the correct size to be a Pcap record header")
        (self.sec, self.usec, self.incl_len, self.orig_len) = struct.unpack(Pcap.RECORD_HEADER_FORMAT, buf)

    def pack(self):
        """
        Pack a PcapRecord into a buffer
        
        :rtype: bytes
        
        """
        if self.sec is None or self.usec is None or self.incl_len is None or self.orig_len is None or self.packet is None:
            raise ValueError("Cannot build record with undefined fields in the payload")

        return struct.pack(Pcap.RECORD_HEADER_FORMAT, self.sec, self.usec, self.incl_len, self.orig_len) + self.packet

    def setCurrentTime(self):
        return self.set_current_time()

    def set_current_time(self):
        """
        Convienece method to set the time of the PCAP record
        
        :rtype: bool 
        """
        currenttime = time.time()
        self.usec = int((currenttime%1)*1e6)
        self.sec = int(currenttime)
        return True

    def __repr__(self):
        return "LEN:{} SEC:{} USEC:{}".format(self.orig_len, self.sec, self.usec)

    def __len__(self):
        return len(self._packet)


class Pcap(object):

    """
    Create a new Pcap object with the specified filename.
    Set the mode to define read, write or append

    :param filename: The PCAP filename
    :type filename: str
    
    
    :Keyword Arguments:
        * *mode* -- r: read w: write a: append
        
    
    Pcap files look like::
    
        -------------- --------------- ---------------- --------------- ---------------- -------
        Global Header | Record Header | Record payload | Record Header | Record payload | .....
        -------------- --------------- ---------------- --------------- ---------------- -------
     
    So after opening the file, iterate through the object to read the records
    
    
    Open a PCAP file for reading. Iterate through the records.

    The pcap can also be treated a list to select the relevant object.
    
    >>> p = Pcap(os.path.join("existing.pcap"))
    >>> print p.network
    0
    >>> for mypcaprecord in p:
    ...    print mypcaprecord.sec
    1111
    >>> import AcraNetwork.SimpleEthernet as SimpleEthernet
    >>> eth = SimpleEthernet.Ethernet()
    >>> eth.unpack(mypcaprecord.payload)
    >>> print eth
    >>> print p[0].sec
    1111
    
    Write a Pcap File
    
    >>> p = Pcap("new.pcap", mode='w')
    >>> p.write_global_header()
    >>> r = PcapRecord()
    >>> r.set_current_time()
    >>> r.payload = eth.pack()
    >>> p.write(r)
    >>> p.close()
            
    """

    GLOBAL_HEADER_FORMAT = '<IhhiIII'
    RECORD_HEADER_FORMAT = '<IIII'
    RECORD_HEADER_SIZE =  struct.calcsize(RECORD_HEADER_FORMAT)
    GLOBAL_HEADER_SIZE = struct.calcsize(GLOBAL_HEADER_FORMAT)

    def __init__(self, filename, **kwargs):

        self.filename = filename #: The filename of the PCAP file
        self.mode = kwargs.get('mode','r') #: The file reading mode
        # Global header fields
        self.magic = 0xa1b2c3d4 #: The magic_number which defines the file format. Leave as is.
        self.versionmaj = 2 #: File format major version. Currently 2
        self.versionmin = 4 #: File format minor version. Currently 4
        self.zone = 0 #: The timezone correction in seconds. 0 = GMT
        self.sigfigs = 0 #: Set to 0
        self.snaplen = 65535 #: snapshot length. Typically unchanged
        self.network  = 1 #: Link-layer header type. http://www.tcpdump.org/linktypes.html
        self._header_written = False

        if 'forreading' in kwargs:
            if kwargs['forreading']  == False:
                self.mode = 'w'
            raise DeprecationWarning("forreading is being deprecated and replaced with the mode argument")

        if self.mode == 'a':
            try:
                self.fopen = open(filename,'ab')
            except:
                raise IOError("Failed to open {} for appending".format(filename))

        elif self.mode == 'r':
            try:
                self.fopen = open(filename,'rb')
                self.filesize = os.path.getsize(filename)
            except Exception as e:
                raise IOError("Failed to open {} for reading {}".format(filename,e))
            else:
                self._read_global_header()

        else:
            self.fopen = open(filename,'wb')
            self.write_global_header()

    def flush(self):
        return self.fopen.flush()

    def readGlobalHeader(self):
        warnings.warn("This method is no longer required", DeprecationWarning)
        pass

    def _read_global_header(self):
        """
        This method will read the pcap global header and unpack it and propogate the relevant attributes
        This should be the first method to call on reading a PCAP file
        
        :rtype: bool
        """

        header = self.fopen.read(Pcap.GLOBAL_HEADER_SIZE)
        (self.magic, self.versionmaj ,self.versionmin, self.zone, self.sigfigs, self.snaplen, self.network) \
            = struct.unpack(Pcap.GLOBAL_HEADER_FORMAT,header)

        return True

    def writeGlobalHeader(self):
        warnings.warn("Replaced with write_global_header", PendingDeprecationWarning)
        self.write_global_header()

    def write_global_header(self):
        """
        Write the global header to a new pcap file
        
        :rtype: None
        """
        if self._header_written:
            return True
        else:
            self._header_written = True
            header = struct.pack(Pcap.GLOBAL_HEADER_FORMAT, self.magic, self.versionmaj, self.versionmin,
                                 self.zone, self.sigfigs, self.snaplen, self.network)
            return self.fopen.write(header)

    def writeARecord(self, pcaprecord):
        warnings.warn("Replace with write", PendingDeprecationWarning)
        self.write(pcaprecord)

    def write(self, pcaprecord):
        """
        Write the supplied pcaprecord to the pcap file
        
        :param pcaprecord: The Pcap Record to write
        :type pcaprecord: PcapRecord
        """

        if not isinstance(pcaprecord, PcapRecord):
            raise Exception("Argument is not of PcapRecord Type")

        self.fopen.write(pcaprecord.pack())

    def close(self):
        """ 
        Close the current pcap file
        
        :rtype: None
        """
        self.fopen.close()

    def readAPacket(self):
        warnings.warn("Replace with the iterator", PendingDeprecationWarning)
        return next(self)

    def __iter__(self):
        return self

    def next(self):

        # read the pcap header to a new object
        pcaprecord = PcapRecord()
        try:
            pcaprecord.unpack(self.fopen.read(Pcap.RECORD_HEADER_SIZE))
        except:
            raise StopIteration

        try:
            pcaprecord.packet = self.fopen.read(pcaprecord.incl_len)
        except:
            raise StopIteration
        else:
            return pcaprecord

    __next__ = next

    def __getitem__(self, item):
        self.fopen.seek(Pcap.GLOBAL_HEADER_SIZE)
        for idx, rec in enumerate(self):
            if idx == item:
                return rec


__author__ = 'DCollins'

import sys
sys.path.append("..")
import os

import unittest
import AcraNetwork.Pcap as pcap
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.iNetX as inetx
import AcraNetwork.MPEGTS as MPEGTS

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


class MPEGTSBasicTest(unittest.TestCase):


    ######################
    # Read a complete pcap file
    ######################

    def test_readFirstMPEGTS(self):
        '''
        Very simple test that reads a pcap file with mpegts packets.
        Takes the first packet in there and decoms the mpegts blocks
        Verifies each block in that first packet
        '''
        p = pcap.Pcap(os.path.join(THIS_DIR, "mpegts_input.pcap"))
        p._read_global_header()
        mypcaprecord = p[0]

        ethpacket = SimpleEthernet.Ethernet()   # Create an Ethernet object
        ethpacket.unpack(mypcaprecord.packet)   # Unpack the pcap record into the eth object
        ippacket =  SimpleEthernet.IP()         # Create an IP packet
        ippacket.unpack(ethpacket.payload)      # Unpack the ethernet _payload into the IP packet
        udppacket = SimpleEthernet.UDP()        # Create a UDP packet
        udppacket.unpack(ippacket.payload)      # Unpack the IP _payload into the UDP packet
        inetxpacket = inetx.iNetX()             # Create an iNetx object
        inetxpacket.unpack(udppacket.payload)   # Unpack the UDP _payload into this iNetX object

        mpegts = MPEGTS.MPEGTS()
        mpegts.unpack(inetxpacket.payload)
        self.assertEqual(mpegts.NumberOfBlocks(),7)
        self.assertFalse(mpegts.contunityerror)
        for packet_index in (list(range(7))):
            if packet_index == 0:
                self.assertEqual(mpegts.blocks[packet_index].pid,0)
                self.assertEqual(mpegts.blocks[packet_index].syncbyte,0x47)
            elif packet_index == 1:
                self.assertEqual(mpegts.blocks[packet_index].pid,4096)
                self.assertEqual(mpegts.blocks[packet_index].syncbyte,0x47)
            else:
                self.assertEqual(mpegts.blocks[packet_index].pid,256)
                self.assertEqual(mpegts.blocks[packet_index].syncbyte,0x47)
        p.close()


    def test_readAllMPEGTS(self):
        '''
        Reads the same mpeg ts file as previously but reads all the data
        in the file and checks for any continuity errors
        :return:
        '''
        p = pcap.Pcap(os.path.join(THIS_DIR, "mpegts_input.pcap"))
        for mypcaprecord in p:

            ethpacket = SimpleEthernet.Ethernet()   # Create an Ethernet object
            ethpacket.unpack(mypcaprecord.packet)   # Unpack the pcap record into the eth object
            ippacket =  SimpleEthernet.IP()         # Create an IP packet
            ippacket.unpack(ethpacket.payload)      # Unpack the ethernet _payload into the IP packet
            udppacket = SimpleEthernet.UDP()        # Create a UDP packet
            udppacket.unpack(ippacket.payload)      # Unpack the IP _payload into the UDP packet
            inetxpacket = inetx.iNetX()             # Create an iNetx object
            inetxpacket.unpack(udppacket.payload)   # Unpack the UDP _payload into this iNetX object
            mpegts = MPEGTS.MPEGTS()
            mpegts.unpack(inetxpacket.payload)
            self.assertEqual(mpegts.NumberOfBlocks(),7)
            self.assertFalse(mpegts.contunityerror)

        p.close()

    def test_stanag(self):
        ts_file = open(os.path.join(THIS_DIR, "stanag_sample.ts"), mode='rb')
        h264_data = MPEGTS.H264()
        self.assertTrue(h264_data.unpack(ts_file.read()))
        ts_file.close()
        self.assertEqual(len(h264_data.nals),5062)
        nal_counts ={}
        for nal in h264_data.nals:
            if not nal.type in nal_counts:
                nal_counts[nal.type] = 1
            else:
                nal_counts[nal.type] += 1
        self.assertEqual(nal_counts[0], 687)
        self.assertEqual(nal_counts[6], 1374)

if __name__ == '__main__':
    unittest.main()



__author__ = 'diarmuid'
import sys
sys.path.append("..")
import os

import unittest
import AcraNetwork.SimpleEthernet as SimpleEthernet
import AcraNetwork.Pcap as pcap
import struct

THIS_DIR = os.path.dirname(os.path.abspath(__file__))

class SimpleEthernetTest(unittest.TestCase):

    ######################
    # Ethernet
    ######################
    def test_DefaultEthernet(self):
        e = SimpleEthernet.Ethernet()
        self.assertEqual(e.dstmac,None)
        self.assertEqual(e.srcmac,None)
        self.assertEqual(e.type, SimpleEthernet.Ethernet.TYPE_IP)
        self.assertEqual(e.payload,None)

    def test_basicEthernet(self):
        '''Create an ethernet frame, then unpack it to a new object'''
        e = SimpleEthernet.Ethernet()
        e.srcmac = 0x001122334455
        e.dstmac = 0x998877665544
        e.type = SimpleEthernet.Ethernet.TYPE_IP
        e.payload = struct.pack("H",0xa)
        ethbuf = e.pack()

        e2  = SimpleEthernet.Ethernet()
        e2.unpack(ethbuf)

        self.assertEqual(e2.dstmac,0x998877665544)
        self.assertEqual(e2.type,SimpleEthernet.Ethernet.TYPE_IP)
        self.assertEqual(e2.srcmac,0x001122334455)

    def test_buildEmptyEthernet(self):
        '''Try and create an empty ethernet frame'''
        e = SimpleEthernet.Ethernet()
        self.assertRaises(ValueError,lambda: e.pack())

    def test_non_def_type(self):
        e = SimpleEthernet.Ethernet()
        e.type = SimpleEthernet.Ethernet.TYPE_PAUSE
        e.dstmac = 0x0180c2000001
        e.srcmac = 0x1
        e.payload = struct.pack(">HH", 0x1, 0x2)
        e2 = SimpleEthernet.Ethernet()
        e2.unpack(e.pack())
        self.assertEqual(e, e2)
        self.assertEqual("SRCMAC=00:00:00:00:00:01 DSTMAC=01:80:C2:00:00:01 TYPE=0X8808", repr(e))

    def test_ethernet_fcs(self):
        e = SimpleEthernet.Ethernet()
        e.type = SimpleEthernet.Ethernet.TYPE_IP
        e.dstmac = 0x0180c2000001
        e.srcmac = 0x1
        e.payload = struct.pack(">HH", 0x1, 0x2)
        ex_fcs = struct.pack(">I", 0x6ed798bf)
        self.assertEqual(e.pack(fcs=True)[-4:], ex_fcs)
        e2 = SimpleEthernet.Ethernet()
        e2.unpack(e.pack(fcs=True), fcs=True)
        self.assertEqual(e, e2)



    ######################
    # IP
    ######################
    def test_defaultIP(self):
        i = SimpleEthernet.IP()
        self.assertRaises(ValueError, lambda : i.pack())

    def test_basicIP(self):
        i = SimpleEthernet.IP()
        i.dstip = "235.0.0.1"
        i.srcip = "192.168.1.1"
        i.payload = struct.pack(">H",0xa5)
        ippayload = i.pack()

        i2 = SimpleEthernet.IP()
        i2.unpack(ippayload)
        self.assertEqual(i2.srcip,"192.168.1.1")
        self.assertEqual(i2.dstip,"235.0.0.1")
        self.assertEqual(i2.payload,struct.pack(">H",0xa5))

    def test_unpackIPShort(self):
        i = SimpleEthernet.IP()
        dummypayload = struct.pack('H',0xa5)
        self.assertRaises(ValueError, lambda : i.unpack(dummypayload))

    ######################
    # UDP
    ######################

    def test_defaultUDP(self):
        u = SimpleEthernet.UDP()
        self.assertRaises(ValueError,lambda :u.pack())

    def test_basicUDP(self):
        u = SimpleEthernet.UDP()
        u.dstport = 5500
        u.srcport = 4400
        u.payload = struct.pack('B',0x5)
        mypacket = u.pack()
        self.assertEqual(mypacket,struct.pack('>HHHHB',4400,5500,9,0,0x5))
        self.assertEqual(repr(u), "SRCPORT=4400 DSTPORT=5500")

    def test_unpackUDPShort(self):
        u = SimpleEthernet.UDP()
        dymmypayload =  struct.pack('H',0xa5)
        self.assertRaises(ValueError,lambda : u.unpack(dymmypayload))


    ######################
    # ICMP
    ######################

    def test_defICMP(self):
        i = SimpleEthernet.ICMP()
        self.assertRaises(ValueError, lambda: i.pack())

    ######################
    # Read a complete pcap file
    ######################
    def test_readUDP(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "test_input.pcap"))
        mypcaprecord = p[0]
        p.close()
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        self.assertEqual(e.srcmac,0x0018f8b84454)
        self.assertEqual(e.dstmac,0xe0f847259336)
        self.assertEqual(e.type,0x0800)
        self.assertEqual(repr(e), "SRCMAC=00:18:F8:B8:44:54 DSTMAC=E0:F8:47:25:93:36 TYPE=0X800")

        # checksum test
        (exp_checksum,) = struct.unpack_from("<H", e.payload, 10)
        ip_hdr_checksum = SimpleEthernet.ip_calc_checksum(e.payload[:10] + e.payload[12:20])
        self.assertEqual(exp_checksum, ip_hdr_checksum)
        i = SimpleEthernet.IP()
        i.unpack(e.payload)
        self.assertEqual(i.dstip, "192.168.1.110")
        self.assertEqual(i.srcip, "213.199.179.165")
        self.assertEqual(i.protocol, 0x6)
        self.assertEqual(i.ttl, 48)
        self.assertEqual(i.flags, 0x2)
        self.assertEqual(i.id, 0x4795)
        self.assertEqual(i.len, 56)
        self.assertEqual(i.version, 4)
        #print i
        self.assertEqual(repr(i), "SRCIP=213.199.179.165 DSTIP=192.168.1.110 PROTOCOL=TCP LEN=56")

    def test_ipv4fragment(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "ipv4frags.pcap"))
        pw = pcap.Pcap(os.path.join(THIS_DIR, "combined.pcap"), mode="w")
        pw.write_global_header()
        r = pcap.PcapRecord()
        e = SimpleEthernet.Ethernet()
        e.unpack(p[0].packet)
        i1 = SimpleEthernet.IP()
        i1.unpack(e.payload)
        e.unpack(p[1].packet)
        p.close()
        i2 = SimpleEthernet.IP()
        i2.unpack(e.payload)
        self.assertEqual(i1.id, i2.id)
        combined = SimpleEthernet.combine_ip_fragments([i1,i2])
        e.payload = combined.pack()
        self.assertEqual(1428, len(combined.pack()))
        r.payload = e.pack()
        pw.write(r)
        pw.close()




    # Write an ICMP
    def test_writeICMP(self):

        p = pcap.Pcap("_icmp.pcap",mode='w')
        p.write_global_header()
        r = pcap.PcapRecord()
        r.setCurrentTime()

        ping_req = SimpleEthernet.ICMP()
        ping_req.type = SimpleEthernet.ICMP.TYPE_REQUEST
        ping_req.code = 0
        ping_req.request_id = 0x100
        ping_req.request_sequence = 123
        ping_req.payload = struct.pack(">32B", *range(32))

        e = SimpleEthernet.Ethernet()
        e.srcmac = 0x001122334455
        e.dstmac = 0x998877665544
        e.type = SimpleEthernet.Ethernet.TYPE_IP

        i = SimpleEthernet.IP()
        i.dstip = "235.0.0.1"
        i.srcip = "192.168.1.1"
        i.protocol = SimpleEthernet.IP.PROTOCOLS["ICMP"]
        i.payload = ping_req.pack()
        e.payload = i.pack()
        r.packet = e.pack()
        p.write(r)
        ping_req.type = SimpleEthernet.ICMP.TYPE_REPLY
        i.payload = ping_req.pack()
        e.payload = i.pack()
        p.close()

        p = pcap.Pcap("_icmp.pcap",mode='w')
        p.write_global_header()
        r = pcap.PcapRecord()
        r.setCurrentTime()
        r.packet = e.pack()
        p.write(r)
        p.close()


    def test_readIPchecksum(self):
        p = pcap.Pcap(os.path.join(THIS_DIR, "inetx_test.pcap"))
        mypcaprecord = p[0]
        e = SimpleEthernet.Ethernet()
        e.unpack(mypcaprecord.packet)
        i = SimpleEthernet.IP()
        self.assertTrue(i.unpack(e.payload))
        p.close()

    @unittest.skip("AFDX broken")
    def test_afdx(self):
        af = SimpleEthernet.AFDX()
        af.type = 1
        af.networkID = 2
        af.equipmentID = 3
        af.interfaceID = 4
        af.vlink = 5
        af.sequencenum = 0
        af.payload = struct.pack(">50B", * range(50))
        self.assertEqual(len(af.pack()), 65)
        b = af.pack()
        af2 = SimpleEthernet.AFDX()
        af2.unpack(b)
        self.assertTrue(af == af2)

    def test_igmp(self):
        #https://www.cloudshark.org/captures/b2e93fcea0c2
        exp_b = struct.pack(">HHHHHHHH", 0x2200, 0xe338, 0x0, 0x1, 0x400, 0x0, 0xefc3, 0x702)
        act_b = SimpleEthernet.IGMPv3.join_groups(["239.195.7.2"])
        self.assertEqual(exp_b, act_b)
        exp_b = struct.pack(">6I", 0x2200f33c, 0x2, 0x2000000, 0xefc30702, 0x2000000, 0xeffffffa)
        act_b = SimpleEthernet.IGMPv3.join_groups(["239.195.7.2", "239.255.255.250"])
        self.assertEqual(exp_b, act_b)

    def test_writeIGMP(self):

        p = pcap.Pcap("_igmp.pcap",mode='w')
        r = pcap.PcapRecord()
        act_b = SimpleEthernet.IGMPv3.join_groups(["239.195.7.2", "239.255.255.250"])

        e = SimpleEthernet.Ethernet()
        e.srcmac = 0x001122334455
        e.dstmac = SimpleEthernet.IGMPv3.MAC_ADDR[SimpleEthernet.IGMPv3.IP_ADDR_JOIN]
        e.type = SimpleEthernet.Ethernet.TYPE_IP

        i = SimpleEthernet.IP()
        i.dstip = SimpleEthernet.IGMPv3.IP_ADDR_JOIN
        i.srcip = "192.168.1.1"
        i.protocol = SimpleEthernet.IP.PROTOCOLS["IGMP"]
        FCS_LEN = 4
        pad_len = 64 - (len(act_b) + SimpleEthernet.Ethernet.HEADERLEN + SimpleEthernet.IP.IP_HEADER_SIZE + FCS_LEN)
        i.payload = act_b + struct.pack(f">{pad_len}B", *([0] * pad_len))
        e.payload = i.pack()
        r.packet = e.pack(fcs=True)
        p.write(r)
        p.close()



if __name__ == '__main__':
    unittest.main()



"""
.. module:: iNET
    :platform: Unix, Windows
    :synopsis: Class to construct and de construct iNET Packets

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"

import struct


class iNETPackage(object):
    """
    The _payload of an iNET packet is multiple Package Headers. This class handles such objects
    """

    PKG_FORMAT = ">IHBBI"
    PKG_FORMAT_LEN = struct.calcsize(PKG_FORMAT)
    REQ_ATTR = ("definitionID", "flags", "timedelta", "payload")
    PAD_BYTE = b'\x00'

    def __init__(self):
        self.definitionID = None  #: Package definition ID
        self.flags = None  #: Package status Flags defined in the MDL document
        self._length = 0
        self.timedelta = 0  #: Package time relative to the parent Message timestamp in nanoseconds
        self.payload = b""  #: The package payload

    def pack(self):
        """
        Pack the packet into a binary format and return as a string

        :rtype: str
        """
        for attr in iNETPackage.REQ_ATTR:
            if getattr(self, attr) is None:
                raise ValueError("Require {} is not defined".format(attr))

        if len(self.payload) % 4 != 0:
            padding = (4-len(self.payload) % 4) * iNETPackage.PAD_BYTE
        else:
            padding = b""

        self._length = iNETPackage.PKG_FORMAT_LEN + len(self.payload)

        return struct.pack(iNETPackage.PKG_FORMAT, self.definitionID, self._length, 0, self.flags,  self.timedelta) + \
               self.payload + padding

    def unpack(self, buf):
        """
        Unpack a raw byte stream to an iNET package.
        Accepts a buffer to unpack as the required argument. Returns the unused buffer so that the unpack method
        can be called repeatedly

        :param buf: The string buffer to unpack
        :type buf: str
        :rtype: str
        """
        (self.definitionID, self._length, _res, self.flags, self.timedelta) = \
            struct.unpack_from(iNETPackage.PKG_FORMAT, buf)
        self.payload = buf[iNETPackage.PKG_FORMAT_LEN:self._length]
        if self._length % 4 != 0:
            padding_len = 4 - (self._length % 4)
        else:
            padding_len = 0

        return buf[self._length + padding_len:]


class iNET(object):
    """ 
    Class to pack and unpack iNET payloads. iNET is standard packet format for use
    in FTI networks. It is usually transmitted in a UDP packet containing parameter data
    acquired from sensors and buses

    The packet structure is defined in IRIG106 Chapter24 http://www.irig106.org/docs/106-17/Chapter24.pdf
    
    Capture a UDP packet and unpack the _payload as an iNET packet

    >>> import socket
    >>>> recv_socket = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)  
    >>> data, addr = recv_socket.recvfrom(2048)
    >>> i = iNET()
    >>> i.unpack(data)
    >>> print i.definition_ID
    6
    >>> print i.packages[0].definitionID
    2

    :type flags: int
    :type type: int
    :type version: int
    :type definition_ID: int
    :type sequence: int
    :type ptptimeseconds: int
    :type ptptimenanoseconds: int
    :type app_fields: list[str]
    :type _payload: str
    :type packages: list[iNETPackage]
    """

    INET_HEADER_FORMAT = '>BBHIIIII'
    INET_HEADER_LENGTH = struct.calcsize(INET_HEADER_FORMAT)
    REQ_ATTR = ("flags", "type", "version", "definition_ID", "sequence", "ptptimeseconds", "ptptimenanoseconds",
                "app_fields", "_payload")

    def __init__(self):
        '''Creator method for an iNET class'''
        self.flags = None  #: Message Flags. Bits 15:8 Reserved.
        self.type = 0  #: Message type
        self._option_wc = 0
        self.version = 1  #: Message version
        self.definition_ID = None  #: Message Definition ID
        self.sequence = None  #: Message Sequence Number
        self._length = None  # Length in bytes including header and _payload
        self.ptptimeseconds = None  #: PTP timestamps in seconds
        self.ptptimenanoseconds = None  #: PTP timestamps in nanoseconds
        self.app_fields = []  #: Optional Application Defined fields.
        self._payload = None  # Payload
        self.packages = []  #: The payload is made up of packages. The packagaes are stored in this attribute list[:class:`iNETPackage`]

    def pack(self):
        """
        Pack the packet into a binary format and return as a string

        :rtype: str|bytes
        """

        _wc_ver = len(self.app_fields) + (self.version << 4)

        self._payload = b""
        for pkg in self.packages:
            self._payload += pkg.pack()

        for attr in iNET.REQ_ATTR:
            if getattr(self, attr) is None:
                raise ValueError("Attribute {} is not defined".format(attr))

        self._length = len(self._payload) + iNET.INET_HEADER_LENGTH + len(self.app_fields) * 4
        packet = struct.pack(iNET.INET_HEADER_FORMAT, _wc_ver, self.type, self.flags,  self.definition_ID,
                             self.sequence, self._length, self.ptptimeseconds, self.ptptimenanoseconds)
        if len(self.app_fields) > 0:
            packet += struct.pack(">{}I".format(len(self.app_fields)), *self.app_fields)

        packet += self._payload

        return packet

    def unpack(self, buf):
        """
        Unpack a raw byte stream to an iNET object
        Accepts a buffer to unpack as the required argument

        :param buf: The string buffer to unpack
        :type buf: str
        :rtype: bool
        """

        if len(buf) < iNET.INET_HEADER_LENGTH:
            raise ValueError("Buffer is too short to be an iNET packet")

        (_wc_ver, _type, self.flags,  self.definition_ID, self.sequence, self._length, self.ptptimeseconds,
         self.ptptimenanoseconds) = struct.unpack_from(iNET.INET_HEADER_FORMAT, buf)
        self.type = _type & 0xF
        self._option_wc = _wc_ver & 0xF
        self.version = (_wc_ver >> 4) & 0xF

        if self._option_wc > 0:
            self.app_fields = list(struct.unpack_from(">{}I".format(self._option_wc), buf[iNET.INET_HEADER_LENGTH:]))

        self._payload = buf[iNET.INET_HEADER_LENGTH + (self._option_wc * 4):]

        package_buf = self._payload
        while len(package_buf) > 0:
            package = iNETPackage()
            package_buf = package.unpack(package_buf)
            self.packages.append(package)

        return True

    def __repr__(self):
        return "MessageDefinitionID={:#0X} Sequence={} Type={} TimeStamp(s)={} TimeStamp(ns)={} OptionWordCount={}".format(
            self.definition_ID, self.sequence, self.type, self.ptptimeseconds, self.ptptimenanoseconds, self._option_wc)

    def __eq__(self, other):
        if not isinstance(other, iNET):
            return False

        for attr in iNET.REQ_ATTR:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __len__(self):
        return len(self.pack())










"""
.. module:: MPEGTS
    :platform: Unix, Windows
    :synopsis: Class to handle MPEG Transport Streams

.. moduleauthor:: Diarmuid Collins <dcollins@curtisswright.com>

"""
__author__ = "Diarmuid Collins"
__copyright__ = "Copyright 2018"
__maintainer__ = "Diarmuid Collins"
__email__ = "dcollins@curtisswright.com"
__status__ = "Production"


import struct
import datetime
import sys

NAL_HEADER = 0x00000001
NAL_HEADER_LEN = 4
NAL_TYPES = { "Unspecified" : 0, "Coded non-IDR": 1, "Coded partition A": 2, "Coded partition B": 3,
              "Coded partition C": 4,
              "Coded IDR": 5, "SEI": 6, "SPS": 7, "PPS": 8, "AUD": 9, "EOSeq": 10, "EOStream": 11,
              "Filler": 12, "SES": 13, "Prefix NAL": 14, "SSPS": 15, "Reserved": 16
              }
# Invert it to go from integer to more useful name
NAL_TYPES_INV = {v: k for k, v in list(NAL_TYPES.items())}
SEI_UNREG_DATA = 5
PY3 = sys.version_info > (3,)


class MPEGTS(object):
    """
    This class handles MPEG Transport Streams.
    https://en.wikipedia.org/wiki/MPEG_transport_stream

    Each transport stream contains 188 byte packets.
    These packets contain either video, audio or metadata information

    :type blocks: list[MPEGPacket]
    
    """

    def __init__(self):

        self.previouscounter = {}
        self.discontinuity = {}
        self.blocks = [] #: List of MPEGPacket objects
        self.contunityerror = False
        self.invalidsync = False
        self.invalidsyncblock = list()

    def unpack(self,buf):
        """
        This method will convert a buffer of bytes into an array of MPEG TS packets
        
        :param buf: The buffer to unpack
        :type buf: str
        
        :rtype: bool
        """

        remainingbytes = 0
        prevcount = self.previouscounter
        block_count = 0
        while remainingbytes < len(buf):
            MpegBlock = MPEGPacket()
            MpegBlock.unpack(buf[remainingbytes:remainingbytes+188])
            block_count += 1
            self.blocks.append(MpegBlock)
            if MpegBlock.invalidsync == True:
                self.invalidsync = True
                self.invalidsyncblock.append(block_count)
            remainingbytes += 188
            if  not MpegBlock.pid in prevcount:
                prevcount[MpegBlock.pid] = MpegBlock.continuitycounter
            elif ((prevcount[MpegBlock.pid]+1) % 16) != MpegBlock.continuitycounter:
                self.contunityerror = True
                self.discontinuity[MpegBlock.pid] = (prevcount[MpegBlock.pid],MpegBlock.continuitycounter)
                prevcount[MpegBlock.pid] = MpegBlock.continuitycounter
            else:
                prevcount[MpegBlock.pid] = MpegBlock.continuitycounter

        self.previouscounter = prevcount
        return True


    def NumberOfBlocks(self):
        """
        How many MPEG blocks in the current TS
        
        :rtype: int 
        """

        return len(self.blocks)

    def FirstCount(self):
        """
        Get the value of the first continuity counter
        
        :rtype: int 
        """
        return self.blocks[0].continuitycounter

    def LastCount(self):
        """
        Get the value of the final continuity counter

        :rtype: int 
        """
        return self.blocks[self.NumberOfBlocks()-1].continuitycounter


class MPEGPacket(object):
    """
    The MPEGPacket is the elementary unit in an MPEG Transport Stream
    It contains an header, in which there's a sync word, continuity counter, and a _payload
    """

    def __init__(self):
        self.packetstrut = struct.Struct('>BHB')
        self.sync = None
        self.pid = None
        self.continuitycounter = None
        self.invalidsync = False
        self.payload = ""

    def unpack(self, buf):
        """
        Converts a buffer into an MPEGTS packet
        
        :param buf: The buffer to unpack into an MPEG Packet
        :type buf: str
        :rtype: bool
        """
        (self.syncbyte, pid_full, counter_full)  = self.packetstrut.unpack_from(buf)
        if self.syncbyte != 0x47:
            self.invalidsync = True

        self.pid = pid_full % 8192
        self.continuitycounter = counter_full % 16
        self.payload = buf[struct.calcsize(">BHB"):]


class H264(object):
    """
    This class will handle H.264 _payload. It can convert a buffer of bytes into an array
    of NALs(https://en.wikipedia.org/wiki/Network_Abstraction_Layer)
    The NALs contain different data, based on their types.
    """

    def __init__(self):
        self.nals = []

    def unpack(self, buf):
        """
        Split the buffer into multiple NALs and store as a H264 object

        :param buf: The buffer to unpack into a H264 object
        :type buf: str
        :rtype: bool
        """
        nal_hdr = struct.pack(">L",NAL_HEADER)
        offsets = string_matching_boyer_moore_horspool(buf,nal_hdr)

        for idx,offset in enumerate(offsets):
            if idx == len(offsets)-1:
                nal_buf = buf[offset:]
            else:
                nal_buf = buf[offset:(offsets[idx+1])]
            nal = NAL()
            nal.unpack(nal_buf)
            nal.offset = offset
            self.nals.append(nal)

        return True


class NAL(object):
    """
    The NAL can be split into the various types of NALs.
    """

    def __init__(self):
        self.type = None
        self.size = None
        self.sei = None
        self.offset = None

    def unpack(self, buf):
        """
        Split the buffer into a NAL object

        :param buf: The buffer to unpack into an NAL
        :type buf: str|bytes
        :rtype: bool
        """

        # First 4 bytes are the NAL_HEADER, then forbidden + type
        (self.type,) = struct.unpack_from(">B", buf, NAL_HEADER_LEN)
        self.type = self.type & 0x1F
        self.size = len(buf)
        if self.type == NAL_TYPES["SEI"]:
            sei = STANAG4609_SEI()
            sei.unpack(buf[(NAL_HEADER_LEN+1):])
            self.sei = sei

    def __len__(self):
        return self.size


class STANAG4609_SEI(object):
    """
    Handle the SEI NAL and more specifically this will handle SEIs defined in 3.14.3.5 of the STANAG standard
    http://www.gwg.nga.mil/misb/docs/nato_docs/STANAG_4609_Ed3.pdf
    """
    def __init__(self):
        self.payloadtype = None
        self.payloadsize = None
        self.unregdata = False
        self.status = None
        self.seconds = None
        self.microseconds = None
        self.nanoseconds = None
        self.time = None
        self.stanag = False

    def unpack(self, buf):
        """
        Unpack the NAL _payload as an STANAG4609_SEI

        :param buf: The buffer to unpack into an STANAG4609_SEI
        :type buf: str
        :rtype: bool
        """

        (self.payloadtype,self.payloadsize) = struct.unpack(">BB",buf[0:2])
        if self.payloadtype == SEI_UNREG_DATA:
            self.unregdata = True
            (sig1,sig2,self.status,ms1,_fix1,ms2,_fix2,ms3,_fix3,ms4,) = struct.unpack_from(">QQBHBHBHBH",buf[2:])
            # combine the time fields (cf  http://www.gwg.nga.mil/misb/docs/nato_docs/STANAG_4609_Ed3.pdf 3.14.3.4 )
            # Verify the signature and if it's good then convert to a time
            if sig1 == 0x4d4953506d696372 and sig2 == 0x6f73656374696d65 and \
                            _fix1 == 0xff and _fix2 == 0xff and _fix3 == 0xff:
                useconds = (ms1<<48)+(ms2<<32)+(ms3<<16)+ms4
                self.seconds = float(useconds)/1.0e6
                self.nanoseconds = (ms3<<16)+ms4
                self.time = datetime.datetime.fromtimestamp(self.seconds)
                self.stanag = True


def string_matching_boyer_moore_horspool(text='', pattern=''):
    """
    Returns positions where pattern is found in text.
    O(n)
    Performance: ord() is slow so we shouldn't use it here
    Example: text = 'ababbababa', pattern = 'aba'
         string_matching_boyer_moore_horspool(text, pattern) returns [0, 5, 7]
    @param text text to search inside
    @param pattern string to search for
    @return list containing offsets (shifts) where pattern is found inside text
    """
    m = len(pattern)
    n = len(text)
    offsets = []
    if m > n:
        return offsets
    skip = []
    for k in range(256):
        skip.append(m)
    for k in range(m-1):
        my = pattern[k]
        if PY3:
            skip[pattern[k]] = m - k - 1
        else:
            skip[ord(pattern[k])] = m - k - 1
    skip = tuple(skip)
    k = m - 1
    while k < n:
        j = m - 1
        i = k
        while j >= 0 and text[i] == pattern[j]:
            j -= 1
            i -= 1
        if j == -1:
            offsets.append(i + 1)
        if PY3:
            k += skip[text[k]]
        else:
            k += skip[ord(text[k])]

    return offsets
